#include <iostream>
#include <sstream>

#include "assert.h"

#include "compile.hh"
#include "symtables.hh"
#include "binding.hh"
#include "basfun.hh"
#include "tristate"

#define CPP_SUBFUNC_NAME(x) (std::string(x)+"_f")
#define CPP_MEMBER_NAME(x) (std::string(x))
#define CPP_UTYPE_TYPENAME(x)  (std::string(x)+"_ut")
#define CPP_UTYPE_NAME(x)  (std::string(x)+"_u")
#define CPP_INT_NAME(x)    (std::string(x)+"_i")
#define CPP_SINGLE_NAME(x) (std::string(x)+"_z")
#define CPP_DOUBLE_NAME(x) (std::string(x)+"_d")
#define CPP_LONG_NAME(x)   (std::string(x)+"_l")
#define CPP_STRING_NAME(x) (std::string(x)+"_s")
#define CPP_ARRAY_NAME(x)  (std::string(x)+"_a")

#define CPP_BYTE_NAME(x)   (std::string(x)+"_b")
#define CPP_UBYTE_NAME(x)  (std::string(x)+"_B")
#define CPP_UINT_NAME(x)   (std::string(x)+"_I")
#define CPP_ULONG_NAME(x)  (std::string(x)+"_U")

static bool DoResumeHandling = false; /* FIXME: Make a commandline option or something */
static bool DoERLhandling    = true;

std::vector<std::string> global_data;

static std::string FunctionReturnVarName;

static const std::pair<std::string, std::string>
    GetTypeBaseNames(const TypeInfo& type, const std::string& name)
{
    switch(type.Type)
    {
        case TypeType::Integer: return std::make_pair("BAS_INT", CPP_INT_NAME(name));
        case TypeType::Single:  return std::make_pair("BAS_SGN", CPP_SINGLE_NAME(name));
        case TypeType::Double:  return std::make_pair("BAS_DBL", CPP_DOUBLE_NAME(name));
        case TypeType::Long:    return std::make_pair("BAS_LNG", CPP_LONG_NAME(name));
        case TypeType::String:
            return type.fixlength
                ? /* FIXME */
                  std::make_pair("BAS_STR", CPP_STRING_NAME(name))
                : std::make_pair("BAS_STR", CPP_STRING_NAME(name));
        case TypeType::UserType:
            return std::make_pair(CPP_UTYPE_TYPENAME(type.utypename), CPP_UTYPE_NAME(name));

        case TypeType::Byte:            return std::make_pair("BAS_BYTE",  CPP_BYTE_NAME(name));
        case TypeType::UnsignedByte:    return std::make_pair("BAS_UBYTE", CPP_UBYTE_NAME(name));
        case TypeType::UnsignedInteger: return std::make_pair("BAS_UINT",  CPP_UINT_NAME(name));
        case TypeType::UnsignedLong:    return std::make_pair("BAS_ULONG", CPP_ULONG_NAME(name));
    }
    /* not reached, but keeps compiler happy */
    assert(!"not_reached");
    return std::make_pair(CPP_UTYPE_TYPENAME(type.utypename), CPP_UTYPE_NAME(name));
}

enum StrCompileDefs
{
    StrCompileMultipurpose,
    StrCompilePrinting,
    StrCompileConstOnly
};
static const std::string CompileStringConst
   (const std::string&str,
    StrCompileDefs defs = StrCompileMultipurpose,
    bool CareAboutNuls = true)
{
    char esc = '"';
    if(defs == StrCompilePrinting && str.size() == 1)
    {
        esc = '\'';
    }
    std::string result;
    result += esc;
    bool contains_nul_char = false;
    unsigned size_minus = 0;
    for(std::size_t a=0; a<str.size(); ++a)
    {
        unsigned char c = str[a];
        if((result.size()-size_minus) >= 77 && esc == '"')
        {
            result += "\"\n\"";
            size_minus = result.size()-1;
        }
        if(c=='\n') { result += "\\n"; continue; }
        if(c=='\r') { result += "\\r"; continue; }
        if(c=='\t') { result += "\\t"; continue; }
        if(c=='\b') { result += "\\b"; continue; }
        if(c=='\a') { result += "\\a"; continue; }
        if(c==esc || c=='\\') { result += '\\'; result += str[a]; continue; }
        if(c < 32 || c >= 127)
        {
            if(c == 0) contains_nul_char = true;
            char Buf[64];
            int width = 1;
            if(a+1 < str.size() && str[a+1]>='0' && str[a+1]<='7') width=3;
            sprintf(Buf, "\\%0*o", width, c);
            result += Buf;
            continue;
        }
        result += str[a];
    }
    result += esc;
    if(esc == '"' && defs != StrCompileConstOnly && contains_nul_char && CareAboutNuls)
    {
        return "std::string(" + result + ", " + std::to_string(str.size()) + ")";
    }
    return result;
}

enum VarDeclType { Init_Var, Param_Ref, Param_Copy, Param_Const, Type_Member, Clear_Var };

static struct VariableDeclarationCompiler
{
    std::string resulting_name;

    void operator() (std::ostream& out, const std::string& name,
                     TypeP typep,
                     VarDeclType decl_type = Init_Var
                    )
    {
        if(!typep)
        {
            std::cout << "/* " << name << " has no type? */";
        }
        std::cout << std::flush;

        assert(typep);
        const TypeInfo& type = *typep;

        const std::pair<std::string, std::string> base = GetTypeBaseNames(type, name);

        switch(type.arraytype)
        {
            case TypeInfo::NotArray:
            {
                if(decl_type == Clear_Var)
                {
                    out << (resulting_name=base.second);
                }
                else
                {
                    bool const_ref = false;
                    if(decl_type == Param_Const)
                        const_ref = !type.is_num();

                    if(const_ref) out << "const ";
                    out << base.first << ' ';
                    if(decl_type == Param_Ref || const_ref) out << "&";
                    out << (resulting_name=base.second);
                }

                if(decl_type == Init_Var)
                {
                    out << " { }";
                }
                else if(decl_type == Clear_Var)
                {
                    switch(type.Type)
                    {
                        case TypeType::Integer:
                        case TypeType::Long:
                            out << " = 0";
                            break;
                        case TypeType::Single:
                        case TypeType::Double:
                            out << " = 0.0";
                            break;
                        case TypeType::String:
                            if(decl_type == Clear_Var)
                                out << ".clear()";
                            break;
                        default:
                            out << " = " << base.first << "()";
                    }
                }
                break;
            }
            case TypeInfo::StaticArray:
            {
                if(decl_type == Clear_Var)
                {
                    expr_var *t = new expr_var(name);
                    t->set_type(typep);
                    stmt_erase e(LexLocation(), t);
                    e.Compile();
                }
                else
                {
                    out << base.first << ' ';
                    if(decl_type == Param_Copy)
                    {
                        fprintf(stderr, "Warning: BYVAL Array parameters not supported\n");
                    }
                    out << (resulting_name=CPP_ARRAY_NAME(base.second));
                    for(std::size_t a=0; a<type.dimensions.size(); ++a)
                        out << (!a ? "[" : "][") << type.dimensions[a].second;
                    out << "]";

                    if(decl_type == Init_Var)
                        out << " = { }";
                }
                break;
            }
            case TypeInfo::DynamicArray:
            {
                if(decl_type == Clear_Var)
                {
                    out << (resulting_name=CPP_ARRAY_NAME(base.second));
                    out << ".Erase()";
                }
                else
                {
                    if(decl_type == Param_Const) out << "const ";
                    out << "BAS_DYNARRAY<" << base.first << ", " << type.dimensions.size() << "> ";
                    if(decl_type == Param_Ref
                    || decl_type == Param_Const) out << '&';
                    out << (resulting_name=CPP_ARRAY_NAME(base.second));

                    if(decl_type == Init_Var)
                    {
                        out << "(empty_struct()";
                        for(std::size_t a=0; a<type.dimensions.size(); ++a)
                        {
                            out << ", "
                                << type.dimensions[a].first
                                << ','
                                << type.dimensions[a].second;
                        }
                        out << ')';
                    }
                }
                break;
            }
            case TypeInfo::ParamArray:
            {
                if(decl_type == Param_Const) out << "const ";
                out << "BAS_MAGICARRAY<" << base.first << ">";
                if(decl_type == Param_Ref
                || decl_type == Param_Const) out << "&";
                out << (resulting_name=CPP_ARRAY_NAME(base.second));
                break;
            }
        }
    }
private:
} CompileVarDeclaration;

static void CompileTypeDefs(const Scope& scope)
{
    for(const auto& t: scope.types)
    {
        UtypeP p = t.second;
        assert(p);

        std::cout << "struct " << CPP_UTYPE_TYPENAME(t.first) << " {\n";

        for(const auto& member: *p)
        {
            std::cout << "  ";
            CompileVarDeclaration(std::cout, CPP_MEMBER_NAME(member.first), member.second,
                Type_Member);
            std::cout << ";\n";
        }

        std::cout << "};\n";
    }
}

static void ProcessScopeVars(const ScopeVarListType& vars, tristate shared_only,
                        bool UtypesOnly = false,
                        VarDeclType decl_type = Init_Var)
{
    for(const auto& v: vars)
    {
        ctxp ctx = v.second;
        assert(ctx);
        if(!shared_only.is_maybe() && ctx->is_shared != shared_only) continue;
        if(ctx->param_index >= 0) continue;

        if(UtypesOnly && ctx->type->Type != TypeType::UserType) continue;

        if(ctx->is_static) std::cout << "static ";

        CompileVarDeclaration(std::cout, v.first, ctx->type, decl_type);
        std::cout << ";\n";
    }
}

static void ProcessScopeVars(const Scope& scope, tristate shared_only, VarDeclType decl_type)
{
    /* When emitting Utypes, ignore non-Utypes (explicitly typed variables) */
    for(unsigned a=0; a<NumTypes; ++a)
        for(unsigned is_array=0; is_array<2; ++is_array)
            ProcessScopeVars(scope.Scopes[a][is_array], shared_only, (TypeType)a == TypeType::UserType, decl_type);
}

static void CompileVars(const Scope& scope, tristate shared_only)
{
    ProcessScopeVars(scope, shared_only, Init_Var);
}

static void EmitClearVars(const Scope& scope)
{
    ProcessScopeVars(scope, maybe, Clear_Var);
}

static void CompileSubFunc(const std::string& name, const SubRoutine& sub, bool decl_only)
{
    std::cout << "static ";
    if(sub.returntype)
    {
        const std::pair<std::string, std::string> base
            = GetTypeBaseNames(*sub.returntype, name);
        /* no "const" here, to not hinder passing the result as reference */
        std::cout << base.first;
    }
    else
        std::cout << "void";

    std::cout << " " << CPP_SUBFUNC_NAME(name) << "(";

    for(std::size_t a=0; a<sub.params.size(); ++a)
    {
        const SubRoutine::ParamSpec& param = sub.params[a];
        if(a) std::cout << ", ";
        CompileVarDeclaration(std::cout,
                              param.name,
                              param.type,
                              param.was_modified
                               ? param.passconvention ==
                                   SubRoutine::ParamSpec::ByVal
                                       ? Param_Copy
                                       : Param_Ref
                               : Param_Const);
    }

    std::cout << ')';

    if(decl_only)
        std::cout << ";\n";
    else
    {
        std::cout << "\n";
    }
}

static void CompileSubDefs(const Scope& scope)
{
    for(const auto& s: scope.subroutines)
    {
        SubFuncP sub = s.second; assert(sub);
        std::cout << "/* DECL OF SUB " << s.first << " */\n" << std::flush;
        CompileSubFunc(s.first, *sub, true);
    }
}

static void CompileCode(stmtp codep)
{
    std::list<stmtp> stack;
    while(codep) { stack.push_front(codep); codep=codep->get_prev(); }

    unsigned discriminator = 0;

    for(std::list<stmtp>::iterator j, i = stack.begin(); i != stack.end(); i=j)
    {
        j=i; ++j;
        codep = *i;

        bool HasErrorHandler = codep->MayErr();

        std::cout << "#line "
            << codep->GetLoc().line
            << ' '
            << CompileStringConst(codep->GetLoc().filename, StrCompileConstOnly)
            << "\n";

        if(HasErrorHandler && DoResumeHandling)
        {
            std::cout << "B_ERR_STMT_BEGIN(); ";
        }

        codep->Compile();

        if(HasErrorHandler && DoResumeHandling)
        {
            if(DoERLhandling)
                std::cout << "B_ERR_STMT_END("
                          << codep->GetLoc().line << ','
                          << codep->GetLoc().col  << ','
                          << CompileStringConst(codep->GetLoc().filename, StrCompileConstOnly)
                          << ',' << discriminator++
                          << ");\n";
            else
                std::cout << "B_ERR_STMT_END_BRIEF();";
        }

        stack.erase(i);
    }
}

static const std::string GenerateLabelName(const std::string& lab)
{
    const Scope& scope = *GetCurrentScope();
    if(scope.labels.find(lab) == scope.labels.end())
    {
        fprintf(stderr, "Error: Label not defined in current scope\n");
        return "L??";
    }

    if(lab[0] >= '0' && lab[0] <= '9') return "L" + lab;

    static std::map<std::pair<const Scope*, std::string>, unsigned> ldata;
    std::pair<const Scope*, std::string> seed(&scope, lab);

    unsigned lnum = 0;
    auto l = ldata.find(seed);
    if(l != ldata.end())
        lnum = l->second;
    else
        lnum = ldata[seed] = ldata.size() + 1;

    return "L" + std::to_string(lnum) + "_" + lab;
}

static void CompileExpr(exprp e,
                        ExprCompileDefs defs=exprdef_nothingparticular)
{
    if(!e) return;

    if(defs == exprdef_cast_to_int)
    {
        TypeP type = e->GetType();
        if(type && !type->is_integer())
        {
            if(expr_number* num = dynamic_cast<expr_number*>(&*e))
            {
                if(num->is_int()) goto NormalExprCompile;
            }
            //std::cout << "/* operand type is " << type->Dump() << " */";

            std::cout << "BasCLNG(";
            defs = exprdef_nothingparticular;
            e->Compile(defs);
            std::cout << ")";
            return;
        }
    }
NormalExprCompile:
    e->Compile(defs);
}

static void VerifyLabels(const Scope& scope)
{
    for(const auto& l: scope.labels)
    {
        if(!l.second.defined)
            fprintf(stderr, "Error: Label %s not defined\n", l.first.c_str());
    }
}

/*
PLAN:
    Define typedefs. - DONE
    Define global variables. -SHARED ONLY DONE
    Declare subroutines. -DONE

    Compile code.

    Compile each function.

NOTES:
    Constants are not defined. They are compiled!

FIXME:
    How to do individually imported globals?
    (SHARED clause within SUBs, works like global in PHP)

*/
void Compile(stmtp main_codep,
             const std::vector<std::string>& global_data)
{
    ScopeP scope = GetRootScope();
    SetCurrentScope(scope);

    VerifyLabels(*scope);
    AnalyzeSubParamUsage();

    if(!DoResumeHandling)
        std::cout << "#define B_NO_ERROR_HANDLING 1\n";

    std::cout << "#include <qblib/qblib.hh>\n";

    std::cout << "constexpr std::size_t DATA_COUNT = " << global_data.size() << ";\n";
    if(!global_data.empty())
    {
        std::cout << "static std::size_t DATA_PTR = 0;\n";
        std::cout << "static const char* const DATA_TABLE[DATA_COUNT] = {\n";
        unsigned line_length = 0;
        for(std::size_t a=0; a<global_data.size(); ++a)
        {
            //std::cout << "// " << a << "\n";
            std::string s = CompileStringConst(global_data[a], StrCompileConstOnly);
            if(line_length > 0 && line_length + s.size() >= 79)
                { std::cout << "\n"; line_length = 0; }
            else if(line_length)
                { std::cout << ' '; ++line_length; }
            std::cout << s;
            if(a+1 < global_data.size()) std::cout << ',';
            line_length += s.size() + 1;
        }
        if(line_length) std::cout << "\n";
        std::cout << "};\n";
    }

    CompileTypeDefs(*scope);

    std::cout << "/* These are global */\n";
    CompileVars(*scope, true); // shared only

    std::cout << "/* These are local to this module */\n";
    std::cout << "template<bool insub>\n";
    std::cout << "static void B_StmtCLEAR(int=-1, int=-1, int=-1);\n";
    CompileSubDefs(*scope);

    std::cout << "namespace {\n";

    std::cout << "/* These are local to main() */\n";
    CompileVars(*scope, false); // local only
    std::cout << "}\n";

    std::cout << "int main(int argc, char**argv)\n";
    std::cout << "{\n";
    std::cout << "B_LibInit(argc, argv);\n";
    std::cout << "B_ERR_FRAME_MAIN_BEGIN();\n";
    CompileCode(main_codep);

    std::cout << "return 0;\n";

    std::cout << "B_ERR_FRAME_MAIN_END();\n";
    std::cout << "}\n";

    /* Compile subs */
    for(const auto& s: scope->subroutines)
    {
        std::cout << "\n\n";

        std::cout << "/* BODY OF SUB " << s.first << " */\n" << std::flush;

        SubFuncP sub = s.second; assert(sub);
        ScopeP subscope = sub->sub_scope;

        if(!&*subscope)
        {
            fprintf(stderr, "Error: SUB %s scope missing\n",
                s.first.c_str());
            continue;
        }

        SetCurrentScope(subscope);

        VerifyLabels(*subscope);

        CompileSubFunc(s.first, *sub, false);
        std::cout << "{\n";

        std::cout << "B_ERR_FRAME_SUB_BEGIN();\n";

        if(sub->returntype)
        {
            CompileVarDeclaration(std::cout, "FunctionReturn", sub->returntype);
            FunctionReturnVarName = CompileVarDeclaration.resulting_name;
            std::cout << ";\n";
        }

        CompileVars(*subscope, false/*, sub->is_static*/); // local only

        CompileCode(sub->code);

        std::cout << "goto subfunc_exit; subfunc_exit:; \n";
        // the goto above supresses a gcc warning.

        if(sub->returntype)
            std::cout << "    return " << FunctionReturnVarName << ";\n";
        else
            std::cout << "    return;\n";

        std::cout << "B_ERR_FRAME_SUB_END();\n";

        std::cout << "}\n";
    }

    std::cout << "\n\n";

    /* Create a B_StmtCLEAR function */
    SetCurrentScope(scope);
    std::cout << "template<bool insub>\n";
    std::cout << "static void B_StmtCLEAR(int ds, int x, int ss)\n{\n";
    EmitClearVars(*scope);
    std::cout << "B_CLEAR(ds,x,ss, insub);\n";
    std::cout << "}\n";
}

void stmt_if::Compile()
{
    if(!thenp && elsep && elseifp.empty())
    {
        std::cout << "if(!("; CompileExpr(condition, exprdef_pure_condition);
        std::cout << ")) {\n"; CompileCode(elsep);
        std::cout << "}\n";
    }
    else
    {
        std::cout << "if("; CompileExpr(condition, exprdef_pure_condition);
        std::cout << ") {\n";
            if(thenp) CompileCode(thenp);
        std::cout << "}";

        for(const auto& e: elseifp)
        {
            std::cout << " else if("; CompileExpr(e.first, exprdef_pure_condition);
            std::cout << ") {\n";
            if(e.second) CompileCode(e.second);
            std::cout << "}";
        }
        if(elsep)
        {
            std::cout << " else {\n";
            CompileCode(elsep);
            std::cout << "}";
        }
        std::cout << "\n";
    }
}
void stmt_doloop::Compile()
{
    std::cout << "{ __label__ do_exit;\n";
    if(precondition)
    {
        std::cout << "while("; CompileExpr(precondition, exprdef_pure_condition); std::cout << ")\n{\n";
    }
    else if(!postcondition)
        std::cout << "for(;;)\n{\n";
    else
        std::cout << "do {\n";

    CompileCode(bodyp);

    if(postcondition)
    {
        if(precondition)
        {
            std::cout << "if("; CompileExpr(postcondition, exprdef_pure_condition);
            std::cout << ") break;";
        }
        else
        {
            std::cout << "} while("; CompileExpr(postcondition, exprdef_pure_condition);
            std::cout << ");";
        }
    }
    else
        std::cout << "}";
    std::cout << " goto do_exit; do_exit:; }\n";
    // the goto above supresses a gcc warning.
}
void stmt_for::Compile()
{
    bool dummyloop = !bodyp;
    /* Dummy loop? It's probably for delay purposes.
     * Let's make a delay then!
     */

    static unsigned ForCounter=0;
    std::string ForTempName = "F_" + std::to_string(ForCounter++);

    bool nonconst_step = true;
    QBtoken step_operator = LESSOREQUAL;

    if(!step)
        nonconst_step = false;
    else
    {
        if(expr_number* num = dynamic_cast<expr_number*> (&*step))
        {
            nonconst_step = false;
            if(num->get_i() < 0) step_operator = GREATEROREQUAL;
        }
    }

    std::pair<std::string, std::string> ForTemp;
    if(nonconst_step)
    {
        ForTemp = GetTypeBaseNames(*step->GetType(), ForTempName);
    }

    if(!dummyloop) std::cout << "{ __label__ for_exit;\n";
    else
    {
        std::cout << "{ unsigned long loop_iterations = 0;";
    }

    if(nonconst_step)
    {
        std::cout << "const " << ForTemp.first << " " << ForTempName << " = ";
        CompileExpr(step);
        std::cout << ";\n";
    }
    std::cout << "for(";
        stmt_let(LexLocation(), var, begin).Compile(); // it appends a semicolon, too.
        if(nonconst_step)
        {
            std::cout << "(" << ForTempName << " < 0 ? (";
            //negative step?
                CompileExpr(var); std::cout << " >= "; CompileExpr(end);
            std::cout << ") : (";
            //positive step?
                CompileExpr(var); std::cout << " <= "; CompileExpr(end);
            std::cout << "))";
        }
        else
        {
            expr_binary(step_operator, var, end).Compile(exprdef_pure_condition);
        }

        std::cout << "; ";

        CompileExpr(var);
        if(!step)
            { std::cout << " += 1 "; }
        else if(nonconst_step)
            { std::cout << " += " << ForTempName; }
        else
            { std::cout << " += "; CompileExpr(step); }
    std::cout << ")\n{\n";
        CompileCode(bodyp);
        if(dummyloop) std::cout << "++loop_iterations;";
    std::cout << "} ";

    if(!dummyloop)
    {
        std::cout << "goto for_exit; for_exit:; }\n";
        // The purpose of this "goto" is just to suppress a compiler warning.
    }
    else
    {
        std::cout << "B_DummyLoop(loop_iterations); }\n";
    }
}
void stmt_let::Compile()
{
    CompileExpr(target);

    if(expr_binary* bin = dynamic_cast<expr_binary*> (&*value))
    {
        std::string op, cast1;
        ExprCompileDefs subdefs = exprdef_nothingparticular;
        switch(bin->token)
        {
            case XOR: op = "^="; break;
            case OR:  op = "|="; break;
            case AND: op = "&="; break;
            case (QBtoken)'+': op = "+="; break;
            case (QBtoken)'-': op = "-="; break;
            case (QBtoken)'*': op = "*="; break;
            case (QBtoken)'/': op = "/="; cast1 = "(BAS_DBL)"; break;
            default: break;
        }
        if(!op.empty())
        {
            /* Check if the first argument in this expr_binary
             * is the same as target, and if it is, act on it
             */
            if(expr_var* tvar = dynamic_cast<expr_var*>(&*target))
                if(expr_var* svar = dynamic_cast<expr_var*>(&*bin->p0))
                {
                    if(tvar->getname() == svar->getname()
                    && *tvar->GetType() == *svar->GetType()
                    && !tvar->isarray() && !svar->isarray())
                    {
                        std::cout << ' ' << op << ' ' << cast1;
                        CompileExpr(bin->p1, subdefs);
                        std::cout << ";\n";
                        return;
                    }
                }
        }
    }

    std::cout << " = ";
    CompileExpr(value);
    std::cout << ";\n";
}

static void CompileCaseTest(exprp expr, const std::string& varname,
                            bool IsSwitchCase)
{
    ExprCompileDefs defs = IsSwitchCase ? exprdef_cast_to_int : exprdef_nothingparticular;
    if(expr_caseor* e = dynamic_cast<expr_caseor*>(&*expr))
    {
        CompileCaseTest(e->a, varname, defs);
        if(!IsSwitchCase) std::cout << " || ";
        CompileCaseTest(e->b, varname, defs);
    }
    else if(expr_caseis* e = dynamic_cast<expr_caseis*>(&*expr))
    {
        std::string op = "==";
        switch(e->token)
        {
            case (QBtoken)'>':    op = ">"; break;
            case (QBtoken)'<':    op = "<"; break;
            case (QBtoken)'=':    op = "=="; break;
            case NOTEQUAL:        op = "!="; break;
            case LESSOREQUAL:     op = "<="; break;
            case GREATEROREQUAL:  op = ">="; break;
            default: assert(!"unexpected caseis token");
        }
        if(IsSwitchCase && op == "==") op="";
        if(IsSwitchCase) std::cout << "case " << op;
        else std::cout << varname << ' ' << op << ' ';
        CompileExpr(e->ref, defs);
        if(IsSwitchCase) std::cout << ":\n";
    }
    else if(expr_caserange* e = dynamic_cast<expr_caserange*>(&*expr))
    {
        if(IsSwitchCase)
        {
            std::cout << "case ";
            CompileExpr(e->min, defs);
            std::cout << " ... ";
            CompileExpr(e->max, defs);
            std::cout << ":\n";
        }
        else
        {
            std::cout << "(" << varname << " >= ";
            CompileExpr(e->min, defs);
            std::cout << " && " << varname << " <= ";
            CompileExpr(e->max, defs);
            std::cout << ')';
        }
    }
    else assert(!"unexpected case type");
}

void stmt_select::Compile()
{
    if(IsEntirelyIntegerCase())
    {
        bool gosubcase = IsEntirelyGosubCase();
        bool had_caseelse = false;

        if(gosubcase)
        {
            std::cout << "{ __label__ gosub_return; B_GOSUB_Push(&&gosub_return); ";
            /* GCC already optimizes switch-case-gotos as jump tables.
             * GOSUB is not much different; we just need to push a return
             * address beforehand. We utilize GCC's existing ability, and
             * turn all GOSUBs into GOTOs.
             */
        }

        std::cout << "switch(";
        CompileExpr(ref, exprdef_cast_to_int);
        std::cout << ")\n{\n";

        for(const auto& c: cases)
        {
            if(gosubcase)
            {
                stmt_labelref* l = dynamic_cast<stmt_labelref*> (&*c.second);
                if(l && l->token == GOSUB) l->token = GOTO; // transform the GOSUBs to GOTOs
            }

            if(dynamic_cast<expr_caseelse*>(&*c.first))
            {
                std::cout << "default:\n";
                CompileCode(c.second);
                had_caseelse = true;
                break;
            }
            CompileCaseTest(c.first, "?", true);
            CompileCode(c.second);
            std::cout << "break;\n";
        }

        if(gosubcase && !had_caseelse)
        {
            std::cout << "default: B_RETURN_TO(gosub_return);\n";
        }

        std::cout << "}\n";
        if(gosubcase)
        {
            std::cout << "gosub_return:; }\n";
        }
    }
    else
    {
        static unsigned SwitchCounter=0;
        std::string SwitchTempName = "S_" + std::to_string(SwitchCounter++);

        std::pair<std::string, std::string> SwitchTemp;
        SwitchTemp = GetTypeBaseNames(*ref->GetType(), SwitchTempName);

        bool first = true;
        std::cout << "{ const " << SwitchTemp.first << " " << SwitchTemp.second
                  << " = "; CompileExpr(ref);
        std::cout << ";\n";

        for(const auto& c: cases)
        {
            if(!first) std::cout << "else ";
            if(dynamic_cast<expr_caseelse*>(&*c.first))
            {
                std::cout << "{\n";
                CompileCode(c.second);
                std::cout << "} ";
                break;
            }
            std::cout << "if(";
            CompileCaseTest(c.first, SwitchTemp.second, false);
            std::cout << ")\n{\n";
            CompileCode(c.second);
            std::cout << "} ";
            first=false;
        }

        std::cout << "}\n";
    }
}

void stmt_print::Compile()
{
    static unsigned UsingCounter=0;
    std::string OutTempName   = "O_" + std::to_string(UsingCounter);
    std::string UsingTempName = "U_" + std::to_string(UsingCounter++);

    if(using_string)
    {
        std::cout << "UsingFormatter "/* << UsingTempName*/ << "(";
        CompileExpr(using_string);
        std::cout << ")\n  .DoParam(";
    }
    if(target)
    {
        std::cout << "GET_STREAM_OUTPUT("; CompileExpr(target);
        std::cout << ')';
    }
    else if(is_lprint)
    {
        std::cout << "GET_PRINTER()";
    }
    else
    {
        std::cout << "GET_STDOUT()";
    }

    enum { nothing, push, param } mode = using_string?param:push;

    for(const auto& p: params)
    {
        switch(p.first)
        {
            case (QBtoken)0: // expression param
                if(using_string)
                {
                    if(mode == push) { std::cout << ";\n"; mode = nothing; }
                    if(mode != param) std::cout << UsingTempName << ".DoParam(" << OutTempName;
                    std::cout << ", ";
                    CompileExpr(p.second);
                    mode = param;
                }
                else
                {
                    if(mode == nothing) { std::cout << OutTempName; mode = push;}
                    std::cout << " << ";
                    CompileExpr(p.second, exprdef_may_produce_char_from_string);
                }
                break;
            case SPC: // SPC function
            {
                if(mode == param) { std::cout << ")"; mode = push; }
                if(mode == push)
                    { std::cout << " << PrintSPCtag{"; CompileExpr(p.second); std::cout << "}"; }
                else
                    { std::cout << OutTempName << ".DoSPC("; CompileExpr(p.second); std::cout << ");\n"; }
                break;
            }
            case TAB: // TAB function
            {
                if(mode == param) { std::cout << ")"; mode = push; }
                if(mode == push)
                    if(p.second)
                        { std::cout << " << PrintTABtag{"; CompileExpr(p.second); std::cout << "}"; }
                    else
                        std::cout << " << PrintNextTABtag{}";
                else
                    if(p.second)
                        { std::cout << OutTempName << ".DoTAB("; CompileExpr(p.second); std::cout << ");"; }
                    else
                        std::cout << OutTempName << ".DoNextTAB();\n";
                break;
            }
            case CRLF: // terminator
            {
                if(mode == param) { std::cout << ")"; mode = push; }
                if(mode == push)
                    std::cout << " << PrintCRLFtag{}";
                else
                    std::cout << OutTempName << ".DoCRLF(); ";
                break;
            }
            default: assert(!"unexpected print operator");
        }
    }
    if(mode == param) std::cout << "); ";
    if(mode == push) std::cout << "; ";
    std::cout << "\n";
}

void stmt_input::Compile()
{
    static unsigned UsingCounter=0;
    std::string InTempName   = "I_" + std::to_string(UsingCounter++);

    std::cout << "{ InputStreamBase& " << InTempName << " = ";
    if(source)
    {
        std::cout << "GET_STREAM_INPUT("; CompileExpr(source);
        std::cout << ')';
    }
    else
    {
        std::cout << "GET_STDIN()";
    }
    std::cout << ";\n";

    if(!is_linput && !source && !prompt) prompt = new expr_string("");

    if(!source && !is_linput)  std::cout << "for(;;) {";

    prompt = new expr_binary((QBtoken)'+', prompt,
                             new expr_string("? "));
    prompt = prompt->optimized();

    if(!source && prompt)
        { std::cout << "GET_STDOUT() << ";
          CompileExpr(prompt,
                      exprdef_may_produce_char_from_string);
          std::cout << ";\n"; }

    if(!is_linput)
        std::cout << InTempName << ".BeginInput();\n";

    for(const auto& p: params)
    {
        std::string method = is_linput ? "ReadLine" : "InputTo";
        std::cout << InTempName << '.' << method
                  << '('; CompileExpr(p); std::cout << ");\n";
    }

    /* Input error from console results in "?Redo from start"
     * Input error from file results in error 62 */

    if(!source && !is_linput)
        std::cout << "if(" << InTempName << ".EndInputOkp()) break; \n"
                  << "GET_STDOUT() << " << InTempName << ".GetRedoMsg(); }\n";
    else
        std::cout << "if(!" << InTempName << ".EndInputOkp()) B_ERROR(62);\n";

    if(!source && !keep_cursor)
        std::cout << "GET_STDOUT().DoCRLF();\n";

    std::cout << "}\n";
}


void stmt_labelref::Compile()
{
    expr_label* e = 0;
    std::string lab;
    std::string lname;
    if(label)
    {
        e = dynamic_cast<expr_label*> (&*label);
        assert(e);
        lab = e->get_label();
        lname = GenerateLabelName(lab);
    }
    switch(token)
    {
        case EXITDO: std::cout << "goto do_exit;\n"; break;
        case EXITFOR: std::cout << "goto for_exit;\n"; break;
        case EXITSUB:
        case EXITFUNCTION:
        case EXITDEF: std::cout << "goto subfunc_exit;\n"; break;
        case RETURN:
            if(label)
                std::cout << "B_RETURN_TO(" << lname << ");\n";
            else
                std::cout << "B_RETURN();\n";
            break;
        case RESUME:
            if(label)
                std::cout << "B_RESUME_TO(" << lname << ");\n";
            else
                std::cout << "B_RESUME();\n";
            break;
        case RESUMENEXT:
            std::cout << "B_RESUME_NEXT();\n";
            break;
        case GOTO:
            std::cout << "goto " << lname << ";\n";
            break;
        case GOSUB:
            std::cout << "B_GOSUB(" << lname << ");\n";
            break;
        case ON_ERROR:
            if(label)
                std::cout << "ON_ERROR_GOTO(" << lname << ");\n";
            else
                std::cout << "ON_ERROR_GOTO_NOWHERE();\n";
            break;
        case ON_PLAY: /* FIXME, not fully considered */
            if(label)
                std::cout << "ON_PLAY_GOTO(" << lname << ");\n";
            else
                std::cout << "ON_PLAY_GOTO_NOWHERE();\n";
            break;
        case ON_KEY: /* FIXME, not fully considered */
            if(label)
                std::cout << "ON_KEY_GOTO(" << lname << ");\n";
            else
                std::cout << "ON_KEY_GOTO_NOWHERE();\n";
            break;
        case ON_TIMER: /* FIXME, not fully considered */
            if(label)
                std::cout << "ON_TIMER_GOTO(" << lname << ");\n";
            else
                std::cout << "ON_TIMER_GOTO_NOWHERE();\n";
            break;
        case RESTORE:
            std::cout << "B_RESTORE(";
            if(!label) std::cout << '0';
            else
            {
                const Scope& scope = *GetCurrentScope();
                auto i = scope.labels.find(lab);
                if(i != scope.labels.end())
                    std::cout << i->second.dataptr;

                std::cout << " /* " << lname << " */";
            }
            std::cout << ");\n";
            break;
        default: assert(!"unexpected stmt_goto token");
    }
}
void stmt_label::Compile()
{
    std::string lname = GenerateLabelName(label);
    std::cout << lname.c_str() << ": ;\n";
}

void stmt_fileopen::Compile()
{
    // B_OPEN(filename,filenumber, mode,     bitmask)
    // B_OPEN(filename,filenumber, mode,     bitmask, reclen)
    // B_OPEN(filename,filenumber, mode_str, bitmask)
    // B_OPEN(filename,filenumber, mode_str, bitmask, reclen)

    // reclen is a runtime parsed integer expression
    // mode_str is a runtime parsed string variable, expect "R","O","I","A","B"
    // mode is:
    //    1 for RANDOM
    //    2 for INPUT and OUTPUT
    //    3 for BINARY
    //    4 for APPEND
    // bitmask is:
    //  &3:
    //    1=reading    (INPUT,         or when overridden by ACCESS: RANDOM,BINARY)
    //    2=writing    (OUTPUT,APPEND, or when overridden by ACCESS: RANDOM,BINARY)
    //    3=read+write (RANDOM,BINARY)
    //    0 = when mode_str is used
    //  &12=
    //    0=shared
    //    4=lock read
    //    8=lock write
    //   12=lock read+write
    std::cout << "B_OPEN("; CompileExpr(filename);
    std::cout << ","; CompileExpr(filenumber);

    if(access)
        switch(mode)
        {
            case RANDOM: std::cout << ",1/*RANDOM*/"; break;
            default:
            case INPUT:
            case OUTPUT: std::cout << ",2/*IN/OUT*/"; break;
            case BINARY: std::cout << ",3/*BINARY*/"; break;
            case APPEND: std::cout << ",4/*APPEND*/"; break;
        }
    else
    {
        std::cout << ","; CompileExpr(mode_str);
    }

    std::cout << "," << (access + 4*locking);
    std::cout << "/*";
        if(access&1) std::cout << "READ ";
        if(access&2) std::cout << "WRITE ";
        if(access&4) std::cout << "LOCKREAD ";
        if(access&8) std::cout << "LOCKWRITE ";
    std::cout << "*/";

    if(reclen)
    {
        std::cout << ","; CompileExpr(reclen);
    }
    std::cout << ");\n";
}

void stmt_subcall::Compile()
{
    expr_userfun f(name, 0, params);
    f.Compile(exprdef_nothingparticular, true);
    std::cout << ";\n";
}
void stmt_read::Compile()
{
    std::cout << "B_READ("; CompileExpr(target);
    std::cout << ");\n";
}

void stmt_redim::Compile()
{
    expr_var* v = dynamic_cast<expr_var*>(&*target);
    assert(v);
    const std::string& name = v->getname();

    VarContext ctx = FindVariableFromScope(name, true);
    if(!ctx) fprintf(stderr, "Error: Cannot find definition of '%s'\n", name.c_str());
    TypeP type = ctx.info->type;

    std::pair<std::string, std::string>
        base = GetTypeBaseNames(*type, BasUCASE(StripTypeSuffix(name)));

    switch(type->arraytype)
    {
        case TypeInfo::NotArray:
        case TypeInfo::StaticArray: std::cout << "/* REDIM eaten by grue? */\n"; break;
        case TypeInfo::DynamicArray:
        case TypeInfo::ParamArray:
            std::cout << CPP_ARRAY_NAME(base.second);
            std::cout << ".Redim(" << dimensions.size();
            for(std::size_t a=0; a<dimensions.size(); ++a)
            {
                std::cout << ", (int)(";
                CompileExpr(dimensions[a].first, exprdef_cast_to_int);
                std::cout << "), (int)(";
                CompileExpr(dimensions[a].second, exprdef_cast_to_int);
                std::cout << ")";
            }
            std::cout << ");\n";
            break;
    }
}
void stmt_erase::Compile()
{
    expr_var* v = dynamic_cast<expr_var*>(&*target);
    assert(v);
    const std::string& name = v->getname();
    std::pair<std::string, std::string>
        base = GetTypeBaseNames(*v->GetType(), BasUCASE(StripTypeSuffix(name)));
    VarContext ctx = FindVariableFromScope(name, true);
    if(!ctx) fprintf(stderr, "Error: Cannot find definition of '%s'\n", name.c_str());
    TypeP type = ctx.info->type;
    switch(type->arraytype)
    {
        case TypeInfo::NotArray:
            std::cout << "/* ERASE eaten by grue? */\n";
            break;
        case TypeInfo::StaticArray:
        {
            if(type->dimensions.size() != 1)
            {
                std::cout << "{ " << base.first << "* E_0p = &";
                std::cout << CPP_ARRAY_NAME(base.second);
                size_t size = 1;
                for(std::size_t a=0; a<type->dimensions.size(); ++a)
                {
                    size *= type->dimensions[a].second;
                    std::cout << "[0]";
                }
                std::cout << ";\nfor(std::size_t E_0=0; E_0 < " << size << "; ++E_0)"
                             "\n\tE_0p[E_0] = " << base.first << "(); }\n";
            }
            else
            {
                for(std::size_t a=0; a<type->dimensions.size(); ++a)
                    std::cout << "for(std::size_t E_" << a << "=0; E_" << a
                              << "<"
                              << type->dimensions[a].second
                              << "; ++E_" << a
                              << ")\n";
                std::cout << CPP_ARRAY_NAME(base.second);
                for(std::size_t a=0; a<type->dimensions.size(); ++a)
                    std::cout << "[E_" << a << ']';
                std::cout << " = " << base.first << "();\n";
            }
            break;
        }
        case TypeInfo::DynamicArray:
        case TypeInfo::ParamArray:
            std::cout << CPP_ARRAY_NAME(base.second);
            std::cout << ".Erase();\n";
            break;
    }
}

static void CompileFunctionCall
    (const std::string& functionname,
     const std::vector<exprp>& params,
     const std::vector<SubRoutine::ParamSpec>& paramdefs,
     bool is_statement
       /* Paramdefs components used:
           * type->arraytype == ParamArray
           * passconvention != ByVal
             && was_modified (if needs to be passed by reference)
           * type (for temporary VarDeclaration)
        */
    )
{
    std::vector<std::pair<bool, std::string> > temps(params.size());
    bool has_temps = false;

    for(std::size_t a=0; a<params.size(); ++a)
    {
        temps[a].first = false;

        /* If the parameter needs to be passed by reference */
        if(!paramdefs[a].type)
        {
            fprintf(stderr, "Error: Type of parameter %u of function %s is undefined\n", (unsigned)a, functionname.c_str());
        }
        else if(paramdefs[a].type->arraytype == TypeInfo::ParamArray)
        {
            if(expr_var* v = dynamic_cast<expr_var*>(&*params[a]))
                if(v->isarray())
                {
                    const TypeInfo& vtype = *v->GetType();

                    if(vtype.arraytype == TypeInfo::ParamArray)
                    {
                        /* Param array does not need special care
                         * when the source is also a param array
                         */
                        continue;
                    }

                    /* It's not a var, so it's a temporary then. */
                    if(!has_temps)
                    {
                        has_temps = true;
                        if(is_statement)
                            std::cout << "{";
                        else
                            std::cout << "({\n"; // GCC extension: statement expressions
                    }

                    static unsigned RefCounter=0;
                    std::string TempName = "A_" + std::to_string(RefCounter++);

                    CompileVarDeclaration(std::cout, TempName,
                                          paramdefs[a].type,
                                          Param_Copy);
                    temps[a].first  = true;
                    temps[a].second = CompileVarDeclaration.resulting_name;

                    std::cout << '(';

                    switch(vtype.arraytype)
                    {
                        case TypeInfo::NotArray: goto eek;
                        case TypeInfo::StaticArray:
                        {
                            std::cout << "&";
                            CompileExpr(params[a], exprdef_nothingparticular);
                            for(std::size_t a=0; a<vtype.dimensions.size(); ++a)
                                std::cout << "[0]";

                            std::cout << ", " << vtype.dimensions.size();
                            for(std::size_t a=0; a<vtype.dimensions.size(); ++a)
                                std::cout << ", " << vtype.dimensions[a].first
                                          << ','  << vtype.dimensions[a].second;
                            break;
                        }
                        case TypeInfo::DynamicArray:
                        case TypeInfo::ParamArray:
                            CompileExpr(params[a], exprdef_nothingparticular);
                            break;
                    }
                    std::cout << ");\n";

                    continue;
                }
        eek:
            fprintf(stderr, "Error: Attempt to pass a non-array to a function that expects array\n");
        }

        if(paramdefs[a].passconvention != SubRoutine::ParamSpec::ByVal
        && paramdefs[a].was_modified)
        {
            /* Check if the parameter to be passed is a temporary. */
            /* If it is, a temporary holder needs to be constructed. */
            if(!dynamic_cast<expr_var*>(&*params[a]))
            {
                /* It's not a var, so it's a temporary then. */
                if(!has_temps)
                {
                    has_temps = true;
                    if(is_statement)
                        std::cout << "{";
                    else
                        std::cout << "({\n"; // GCC extension: statement expressions
                }

                static unsigned RefCounter=0;
                    std::string TempName = "R_" + std::to_string(RefCounter++);

                CompileVarDeclaration(std::cout, TempName,
                                      paramdefs[a].type,
                                      Param_Copy);
                temps[a].first  = true;
                temps[a].second = CompileVarDeclaration.resulting_name;
                std::cout << '=';
                CompileExpr(params[a], exprdef_nothingparticular);
                std::cout << ';';
            }
        }
    }

    std::cout << functionname;

    if(functionname == "B_StmtCLEAR")
    {
        // special case
        bool in_sub = &*GetCurrentScope() != &*GetRootScope();

        if(in_sub)
            std::cout << "<true>";
        else
            std::cout << "<false>";
    }

    std::cout << '(';
    for(std::size_t a=0; a<params.size(); ++a)
    {
        if(a) std::cout << ", ";

        if(temps[a].first)
            std::cout << temps[a].second;
        else
            CompileExpr(params[a], exprdef_nothingparticular);
    }
    std::cout << ')';

    if(has_temps)
    {
        if(is_statement)
            std::cout << "; }";
        else
            std::cout << ";\n})";
    }
}

static const std::vector<SubRoutine::ParamSpec>
    CreateParamDefsFromPattern(const std::string& pattern,
        std::vector<exprp>& params)
{
    std::vector<SubRoutine::ParamSpec> paramdefs(params.size());

    unsigned paramno=0;
    const char* defs = pattern.c_str();
    while(*defs)
    {
        #define EP() do { \
            while(paramno >= params.size()) params.push_back((exprp)0); \
            if(paramno >= paramdefs.size()) paramdefs.resize(paramno+1); } while(0)
        #define PTYPE() do { \
            if(params[paramno]) \
                paramdefs[paramno].type = params[paramno]->GetType(); \
            else \
                fprintf(stderr, "Warning: param %u is empty\n", paramno); \
            } while(0)
        #define SetByVal() do { \
            paramdefs[paramno].was_modified = false; \
            paramdefs[paramno].passconvention = SubRoutine::ParamSpec::ByVal; } while(0)
        #define SetByRef() do { \
            paramdefs[paramno].was_modified = true; \
            paramdefs[paramno].passconvention = SubRoutine::ParamSpec::ByRef; } while(0)
        #define InsInt(v) params.insert(params.begin()+paramno, new expr_number(v))
        #define InsStr(s) params.insert(params.begin()+paramno, new expr_string(s))

        if(paramno+1 > params.size()) params.push_back((exprp)0);
        switch(*defs++)
        {
            case 'n': case 'x':
                EP(); if(!params[paramno]) { params[paramno]=new expr_number(-1); }
                //passthru
            case 'N': case 'S': case 'X': // number or string
              GotS:
                EP(); PTYPE(); SetByVal(); ++paramno;
                break;
            case 's':
                EP(); if(!params[paramno]) { params[paramno]=new expr_string(""); }
                goto GotS;
            case 'p':
                EP(); if(!params[paramno]) { params[paramno]=new expr_number(0);
                                             InsInt(0); InsInt(-1); }
                //passthru
            case 'P': // step coordinate
                EP(); PTYPE(); SetByVal(); ++paramno;
                EP(); PTYPE(); SetByVal(); ++paramno;
                EP(); PTYPE(); SetByVal(); ++paramno;
                break;
            case 'c':
                EP(); if(!params[paramno]) { params[paramno]=new expr_number(0);
                                             InsInt(-1); }
                //passthru
            case 'C': // coordinate
                EP(); PTYPE(); SetByVal(); ++paramno;
                EP(); PTYPE(); SetByVal(); ++paramno;
                break;
            case 'V': // lvalue
                EP(); PTYPE(); SetByRef(); ++paramno;
                break;
            case 'A': // array target or source (GET, PUT, PALETTE USING)
                EP(); PTYPE(); SetByRef(); ++paramno;
                break;
            case 'L': // label
            case 'l':
                assert(!"Label params should only occur in stmt_labelref");
            default:
                fprintf(stderr, "Internal Error: pattern '%s': unknown char '%c'\n", pattern.c_str(), *defs);
        }

        #undef EP
        #undef PTYPE
        #undef SetByVal
        #undef SetByRef
        #undef InsInt
        #undef InsStr
    }
    /*
    fprintf(stderr, "params.size=%u, paramno=%u, paramdefs.size=%u, pattern=%s\n",
        (unsigned)params.size(), paramno, (unsigned)paramdefs.size(), pattern.c_str());
    */
    return paramdefs;
}

void stmt_lib_stmt::Compile()
{
    const char* funcname = 0;
    #define tokcase(t, p) \
        case t: funcname = "B_Stmt"#t; goto DoParams;
    switch(token)
    {
/* #include "libdata-stmt-label.inc" - handled separately */
#include "libdata-stmt-stmt.inc"

        case CLOSE: funcname = "B_StmtCLOSE"; goto DoParams;
        case FIELD: funcname = "B_StmtFIELD"; goto DoParams;

        default:
            fprintf(stderr, "Internal Error: Unknown stmt_lib_stmt token: %s\n", GetTokenName(token).c_str());
            assert(!"unknown stmt_lib_stmt token");
    }
    return;
DoParams:
    std::vector<exprp> plist = params;
    std::vector<SubRoutine::ParamSpec> paramdefs
        = CreateParamDefsFromPattern(pattern, plist);

    CompileFunctionCall(
        funcname,
        plist,
        paramdefs,
        true
    );
    #undef tokcase
    std::cout << ";\n";
}

void expr_string::Compile(ExprCompileDefs defs)
{
    std::cout << CompileStringConst
        (this->value,
           defs==exprdef_may_produce_char_from_string
           ? StrCompilePrinting
           : StrCompileMultipurpose);
}
void expr_number::Compile(ExprCompileDefs defs)
{
    if(is_int())
    {
        std::cout << get_i();
        if(type->is_unsigned()) std::cout << 'u';
        if(type->Type == TypeType::Long
        || type->Type == TypeType::UnsignedLong) std::cout << 'l';
    }
    else
    {
        std::cout.precision(20);
        std::cout << get_f();
        /* FIXME: Prevent integer overflows */
    }
}
void expr_unary::Compile(ExprCompileDefs defs)
{
    std::cout << "(";
    ExprCompileDefs subdefs = exprdef_nothingparticular;
    switch(token)
    {
        case NOT:
            if(defs == exprdef_pure_condition) { std::cout << "!"; subdefs=defs; break; }
            std::cout << "~";
            subdefs = exprdef_cast_to_int;
            break;
        case (QBtoken)'-':
            std::cout << "-";
            break;
        default: assert(!"unexpected expr_unary token");
    }
    CompileExpr(p, subdefs);
    std::cout << ')';
}

static void OptimizeStringComparison(exprp& param, int& offset, int& sizelimit)
{
    expr_builtin_string* e = dynamic_cast<expr_builtin_string*>(&*param);
    if(!e) return;

    expr_number* n1 = dynamic_cast<expr_number*> (&*e->GetParam(1));
    expr_number* n2 = dynamic_cast<expr_number*> (&*e->GetParam(2));

    if(e->GetToken() == LEFT_S && n1)
    {
        OptimizeStringComparison(e->GetParamRef(0), offset, sizelimit);

        int left = n1->get_i(); // - offset;
        if(sizelimit == -1 || left < sizelimit) sizelimit = left;
    }
    else if(e->GetToken() == MID_S && n1 && (!e->GetParam(2) || n2))
    {
        OptimizeStringComparison(e->GetParamRef(0), offset, sizelimit);

        int mid_pos = n1->get_i() - 1, mid_len = n2 ? n2->get_i() : -1;

        offset += mid_pos;
        if(sizelimit != -1) sizelimit = std::max(0, sizelimit - mid_pos);

        if(mid_len != -1)
            if(sizelimit == -1 || mid_len < sizelimit) sizelimit = mid_len;
    }
    else return;

    param = e->GetParam(0);
}

void expr_binary::Compile(ExprCompileDefs defs)
{
    std::string fucast, funame, cast0, op, cast1, finish;
    ExprCompileDefs subdefs = exprdef_nothingparticular;

    switch(token)
    {
        case XOR:
            op = "^";
        FinishInteger:
            subdefs = exprdef_cast_to_int;
            break;
        case qbcSHL: op = "<<"; goto FinishInteger;
        case qbcSHR: op = ">>"; goto FinishInteger;
        case OR:
            if(defs == exprdef_pure_condition && IsComparison())
                { op = "||"; subdefs=defs; break; }
            op = "|"; goto FinishInteger;
        case AND:
            if(defs == exprdef_pure_condition && IsComparison())
                { op = "&&"; subdefs=defs; break; }
            op = "&"; goto FinishInteger;
        case (QBtoken)'>':
            op = ">";
        FinishCompare:
            if(defs == exprdef_pure_condition)
                subdefs = defs;
            else
            {
                fucast = "(";
                finish = ") ? -1 : 0";
            }
            break;
        case (QBtoken)'<':    op = "<"; goto FinishCompare;
        case (QBtoken)'=':    op = "=="; goto FinishCompare;
        case NOTEQUAL:        op = "!="; goto FinishCompare;
        case LESSOREQUAL:     op = "<="; goto FinishCompare;
        case GREATEROREQUAL:  op = ">="; goto FinishCompare;
        case (QBtoken)'+':    op = "+"; break;
        case (QBtoken)'-':    op = "-"; break;
        case (QBtoken)'\\':   op = "/"; subdefs = exprdef_cast_to_int; break;
        case (QBtoken)'*':    op = "*"; break;
        case (QBtoken)'/':    op = "/"; cast1 = "(BAS_DBL)"; break;
        case (QBtoken)'^':    funame = "std::pow"; op=","; finish=""; break;
        case MOD:             funame = "BasMODop"; op=","; finish=""; break;
        default: assert(!"unexpected expr_binary token");
    }

    if(p0->GetType()->is_string()
    && p1->GetType()->is_string())
    {
        int gt=0, eq=-1;
        switch(token)
        {
            case (QBtoken)'>':   gt=1, eq=0; break;
            case (QBtoken)'<':   gt=-1,eq=0; break;
            case (QBtoken)'=':   gt=0, eq=1; break;
            case NOTEQUAL:       gt=0, eq=0; break;
            case LESSOREQUAL:    gt=-1,eq=1; break;
            case GREATEROREQUAL: gt=1, eq=1; break;
            default:  goto not_string_compare;
        }
        if(eq != -1)
        {
            /*
            Due to optimization, it shouldn't be possible for both
            strings to be constants. If the lefthand side is a constant,
            swap the operands so that the lefthand side is an actual
            object.
            */
            if(p0->is_const())
            {
                std::swap(p0, p1);
                gt = -gt;
            }

            /* Compile into BAS_STR().compare() */

            std::cout << fucast << funame << "(";

            int lhs_offset = 0, lhs_size = -1;
            int rhs_offset = 0, rhs_size = -1;
            OptimizeStringComparison(p0, lhs_offset, lhs_size);
            OptimizeStringComparison(p1, rhs_offset, rhs_size);

            if(lhs_offset == 0 && lhs_size == -1
            && (rhs_offset != 0 || rhs_size != -1))
            {
                if(!p1->is_const())
                {
                    /* Swap them again to get less parameters */
                    std::swap(p0, p1);
                    gt = -gt;
                    std::swap(lhs_offset, rhs_offset);
                    std::swap(lhs_size,   rhs_size);
                }
            }

            op = "";
            if(gt != 0)            op += (gt > 0  ? '>' : '<');
            else                   op += (eq != 0 ? '=' : '!');
            if(eq != 0 || gt == 0) op += '=';

            if(expr_string* s = dynamic_cast<expr_string*> (&*p0))
            {
                if(lhs_size == -1 || s->get().size() < lhs_size)
                    lhs_size = s->get().size();
            }
            if(expr_string* s = dynamic_cast<expr_string*> (&*p1))
            {
                if(rhs_size == -1 || s->get().size() < rhs_size)
                    rhs_size = s->get().size();
            }


            std::cout << cast0;

            if(expr_string* s = dynamic_cast<expr_string*> (&*p0))
                std::cout << CompileStringConst(s->get(), StrCompileConstOnly, false);
            else
                CompileExpr(p0, subdefs);

            std::cout << ".compare(";
                if(lhs_offset != 0
                || lhs_size   != -1
                || rhs_offset != 0
                || rhs_size   != -1) std::cout << lhs_offset << ", ";
                if(lhs_size  != -1) std::cout << lhs_size << ", ";
                else if(rhs_offset != 0
                     || rhs_size   != -1)
                     {
                         std::cout << "std::string::npos, ";
                     }

                std::cout << cast1;

                bool rhs_const = false;
                if(expr_string* s = dynamic_cast<expr_string*> (&*p1))
                {
                    std::cout << CompileStringConst(s->get(), StrCompileConstOnly, false);
                    rhs_const = true;
                }
                else
                    CompileExpr(p1, subdefs);

                if(rhs_offset != 0 || (!rhs_const && rhs_size  != -1))
                    std::cout << ", " << rhs_offset;

                if(rhs_size  != -1) std::cout << ", " << rhs_size;
            std::cout << ")" << op << "0";

            std::cout << finish << ")";
            return;
        }
    not_string_compare:;
    }

    std::cout << fucast << funame << "("
              << cast0; CompileExpr(p0, subdefs);
    std::cout << ' ' << op << ' '
              << cast1; CompileExpr(p1, subdefs);
    std::cout << finish
              << ')';
}

void expr_userfun::Compile(ExprCompileDefs defs)
{
    Compile(defs, false);
}

void expr_userfun::Compile(ExprCompileDefs defs, bool is_statement)
{
    std::string basename = BasUCASE(StripTypeSuffix(name));
    SubFuncP subp = FindSubFunc(basename);
    if(!subp)
    {
        fprintf(stderr, "Error: Sub/Func %s (%s) not found\n", name.c_str(), basename.c_str());
        return;
    }
    SubRoutine& sub = *subp;
    const std::vector<SubRoutine::ParamSpec>& paramdefs = sub.params;

    if(params.size() != paramdefs.size())
    {
        fprintf(stderr, "Error: Function %s is given %u arguments, expects only %u\n",
            name.c_str(), (unsigned)params.size(), (unsigned)paramdefs.size());
        return;
    }

    CompileFunctionCall(
        CPP_SUBFUNC_NAME(basename),
        params,
        paramdefs,
        is_statement
    );
}

static void TranslateBuiltinFunc(
    QBtoken token,
    std::vector<exprp>& params,
    const std::string& pattern,
    ExprCompileDefs defs)
{
    const char* funcname = 0;
    #define tokcase(t, p) \
        case t: funcname = "B_Func"#t; goto DoParams;
    switch(token)
    {
#include "libdata-expr_numeric.inc"
#include "libdata-expr_string.inc"

        default:
            fprintf(stderr, "Internal Error: Unknown expr token: %s\n", GetTokenName(token).c_str());
            assert(!"unknown expr token");
    }
    return;
DoParams:

    // Check if we can do special optimizations
    expr_number* p0n = params.size() >= 1 ? dynamic_cast<expr_number*>(&*params[0]) : nullptr;
    expr_number* p1n = params.size() >= 2 ? dynamic_cast<expr_number*>(&*params[1]) : nullptr;
    expr_string* p1s = params.size() >= 2 ? dynamic_cast<expr_string*>(&*params[1]) : nullptr;

    switch(token)
    {
        case SPACE_S:
            if(p0n && p0n->get_i() >= 1)
                { std::cout << "std::string(" << p0n->get_i() << ", ' ')"; return; }
            break;
        case STRING_S:
            if(p0n && p0n->get_i() >= 1
            && p1n && p1n->get_i() >= 0 && p1n->get_i() <= 255)
                { std::cout << "std::string(" << p0n->get_i() << ", char(" << p1n->get_i() << "))"; return; }
            if(p0n && p0n->get_i() >= 1
            && p1s && p1s->get().size() >= 1)
                { std::cout << "std::string(" << p0n->get_i() << ", char(" << (int)p1s->get()[0] << "))"; return; }
            break;
        default: ;
    }


    std::vector<exprp> plist = params;
    std::vector<SubRoutine::ParamSpec> paramdefs
        = CreateParamDefsFromPattern(pattern, plist);
    CompileFunctionCall(
        funcname,
        plist,
        paramdefs,
        false
    );
    #undef tokcase
}


void expr_builtin_numeric::Compile(ExprCompileDefs defs)
{
    TranslateBuiltinFunc(token, params, pattern, defs);
}
void expr_builtin_string::Compile(ExprCompileDefs defs)
{
    TranslateBuiltinFunc(token, params, pattern, defs);
}

void expr_lubound::Compile(ExprCompileDefs defs)
{
    bool kind_u = token==UBOUND;

    const char* fname = kind_u?"UBOUND":"LBOUND";
    const char* bname = kind_u?"Ubound":"Lbound";

    expr_var* v = dynamic_cast<expr_var*>(&*var);
    if(!v) { ae:fprintf(stderr, "Error: %s must only be used on arrays\n",fname); return; }
    if(!v->isarray()) goto ae;

    const std::string& name = v->getname();
    VarContext ctx = FindVariableFromScope(name, true);
    if(!ctx) fprintf(stderr, "Error: Cannot find definition of '%s'\n", name.c_str());
    TypeP type = ctx.info->type;

    switch(type->arraytype)
    {
        case TypeInfo::NotArray: goto ae;
        case TypeInfo::StaticArray:
        {
            if(expr_number* e = dynamic_cast<expr_number*>(&*num_index))
            {
                long index = e->get_i();
                if(index < 1 || index > type->dimensions.size())
                    fprintf(stderr, "Error: Array subscript out of range\n");
                else
                    CompileExpr(new expr_number(
                        type->dimensions[index-1].first
                      + (kind_u ? type->dimensions[index-1].second-1 : 0) ));
            }
            else
            {
                std::cout << "({ int bound=0; switch(";
                CompileExpr(num_index, exprdef_cast_to_int);
                std::cout << ") {\n";
                for(std::size_t a=0; a<type->dimensions.size(); ++a)
                    std::cout << "case " << (a+1) << ": bound="
                    << (type->dimensions[a].first
                      + (kind_u ? type->dimensions[a].second-1 : 0))
                    <<  "; break;\n"
                             "default: B_ERROR(9); \n" /* Subscript out of range */
                             "} bound; })";
            }
            break;
        }
        case TypeInfo::DynamicArray:
        case TypeInfo::ParamArray:
        {
            std::pair<std::string, std::string>
                base = GetTypeBaseNames(*type, BasUCASE(StripTypeSuffix(name)));

            std::cout << CPP_ARRAY_NAME(base.second)
                      << '.' << bname
                      << '(';
            CompileExpr(num_index, exprdef_cast_to_int);
            std::cout << ')';
            break;
        }
    }
}

// SPECIALS
void expr_label::Compile(ExprCompileDefs defs)    { assert(!"not reached"); }
void expr_caseelse::Compile(ExprCompileDefs defs) { assert(!"not reached"); }
void expr_caseis::Compile(ExprCompileDefs defs)   { assert(!"not reached"); }
void expr_caserange::Compile(ExprCompileDefs defs){ assert(!"not reached"); }
void expr_caseor::Compile(ExprCompileDefs defs)   { assert(!"not reached"); }

// LVALUES
void expr_var::Compile(ExprCompileDefs defs)
{
    TypeP basetype = type;
    if(defs != exprdef_membername)
    {
        VarContext ctx = FindVariableFromScope(name, isarray());
        if(!ctx)
        {
            fprintf(stderr, "Error: Cannot find definition of '%s'\n", name.c_str());
        }
        assert(ctx);
        assert(ctx.info);
        basetype = ctx.info->type;
    }
    assert(basetype);

    std::pair<std::string, std::string>
        base = GetTypeBaseNames(*basetype, BasUCASE(StripTypeSuffix(name)));

    if(defs==exprdef_membername)
        base.second = CPP_MEMBER_NAME(base.second);

    switch(basetype->arraytype)
    {
        case TypeInfo::NotArray:
            std::cout << base.second;
            break;
        case TypeInfo::StaticArray:
            std::cout << CPP_ARRAY_NAME(base.second);
            for(std::size_t a=0; a<indexes.size(); ++a)
            {
                std::cout << '[';
                exprp index = indexes[a];
                if(basetype->dimensions[a].first != 0)
                {
                    index = new expr_binary
                        ((QBtoken)'-', index, new expr_number(basetype->dimensions[a].first));
                    index = index->optimized();
                }
                CompileExpr(index, exprdef_cast_to_int);
                std::cout << ']';
            }
            break;
        case TypeInfo::DynamicArray:
        case TypeInfo::ParamArray:
            std::cout << CPP_ARRAY_NAME(base.second);
            if(!array_as_whole)
            {
                std::cout << ".Index(" << indexes.size();
                for(std::size_t a=0; a<indexes.size(); ++a)
                {
                    std::cout << ", ";
                    CompileExpr(indexes[a], exprdef_cast_to_int);
                }
                std::cout << ')';
            }
            break;
    }
    if(hasmember())
    {
        std::cout << '.';
        member->Compile(exprdef_membername); // not CompileExpr here
    }
}
void expr_returnvalueholder::Compile(ExprCompileDefs defs)
{
    std::cout << FunctionReturnVarName;
}
