/* A (hopefully) light-weight replacement for boost::spirit */

/*
 * The real motivation behind this creation, which is literally a case
 * of "reinventing the wheel", is because I couldn't compile QBC with
 * Spirit because the debugging information took hundreds of megabytes,
 * thus exhausting the diskspace from my /tmp.
 * In this parser, the symbols are deliberately terse to prevent that
 * from happening.
 * I have tried to maintain compatibility to Spirit where reasonably
 * possible. Noticeably, the parser and scanner are profoundly different.
 *
 *  --Bisqwit
 */

template<typename IteratorT>
class scanner
{
public:
    typedef IteratorT iterator;
    typedef __typeof__(*iterator()) value_type;
    typedef typename iterator::LocType LocT;
    typedef typename iterator::InfoType InfoT;

    scanner(const iterator& b, const iterator& e)
        : cur(b), last(e) { }

    bool eof() const { return cur==last; }
    const value_type get()
    {
        value_type v = *cur;
        info = cur.GetInfo();
        ++cur;
        return v;
    }
    const iterator& pos() const { return cur; }
    void revert(iterator i) { cur=i; if(!eof()) info = cur.GetInfo(); }

    template<typename ActionT>
    void act(const ActionT& act, const iterator& begin, const iterator& end)
    {
#if 0
        dprintf(dp_lexer, "act:act(%p)begin(%p)end(%p)\n",
            (void*)(&act), (void*)begin, (void*)end);
#endif
        act(begin,end);
    }

    const LocT GetLoc() const { return cur.GetLoc(); }
    const InfoT& GetInfo() const { return info; }

private:
    iterator cur,last;
    InfoT info;
};

template<typename ParserT, typename ActionT>
class Actn; //  forward declaration

template<typename ParserT>
class Expecter;

template<typename DerivedT>
class parser
{
public:
    typedef DerivedT embed_t;

    DerivedT& derived()
    {
        return *static_cast<DerivedT*>(this);
    }

    DerivedT const& derived() const
    {
        return *static_cast<DerivedT const*>(this);
    }

    template<typename ActionT>
    Actn<DerivedT,ActionT> operator[] (const ActionT& func) const
    {
        return Actn<DerivedT,ActionT> (derived(), func);
    }

    Expecter<DerivedT> expected() const
    {
        return Expecter<DerivedT> (derived());
    }

    template<typename ScannerT>
    bool parse(ScannerT& scan) const
    {
        return derived().parse(scan);
    }

    /* */
    //static const std::string what() { return "something"; }

protected:
    template<class CharT>
    static const std::basic_string<CharT>
        what2join(const std::basic_string<CharT>&a,
                  const std::basic_string<CharT>&b,
                  const CharT* const sep)
    {
        if(a.empty() || b.empty()) return a + b;
        return a + sep + b;
    }
};

struct NopActor {
    template<typename I>
    inline void operator()(const I&,const I&) const {}
    inline void operator()() const {}
};

class InvalidateAction
{
    const std::string msg;
public:
    InvalidateAction() : msg()                      { }
    InvalidateAction(const char*s) : msg(s)         { }
    InvalidateAction(const std::string& s) : msg(s) { }

    const std::string what() const { return msg.empty()?"Parse error":msg; }
};

template<typename ParserT, typename ActionT>
class Actn: public parser<Actn<ParserT,ActionT> >
{
    typedef parser<Actn<ParserT,ActionT> > base_t;
    const typename ParserT::embed_t opt;
    ActionT act;
public:
    Actn(const ParserT& p, const ActionT& a)
        : base_t(),opt(p.derived()),act(a) { }

    template<typename ScannerT>
    bool parse(ScannerT& scan) const
    {
        typename ScannerT::iterator it = scan.pos();
        if(!opt.parse(scan)) return false;
        try
        {
            scan.act(act, it, scan.pos());
        }
        catch(const InvalidateAction& inv)
        {
            scan.revert(it);
            yyerror(inv.what(), it.GetLoc());
            return false;
        }
        return true;
    }

    inline const std::string what() const { return opt.what(); }
};

template<typename ParserT>
class Expecter: public parser<Expecter<ParserT> >
{
    typedef parser<Expecter<ParserT> > base_t;
    const typename ParserT::embed_t opt;
public:
    Expecter(const ParserT& p) : base_t(), opt(p.derived()) { }

    template<typename ScannerT>
    bool parse(ScannerT& scan) const
    {
        if(opt.parse(scan)) return true;
        if(scan.eof()) yyerror("Expected: "+opt.what());
        else yyerror("Expected: "+opt.what(), scan.GetLoc());
        return false;
    }

    inline const std::string what() const { return opt.what(); }
};

///////////////

#include "tokens.hh"

#define GenBinaryOpsTokenCast(cls, op, src) \
    template<typename A> inline const cls<A,TKn> \
    operator op (const parser<A>& a, src b) { return cls<A,TKn>(a, ch_p(b)); } \
    template<typename B> inline const cls<TKn,B> \
    operator op (src a, const parser<B>& b) { return cls<TKn, B>(ch_p(a), b); }

#define GenBinaryOps(cls, op) \
    template<typename A, typename B> inline const cls<A,B> \
    operator op (const parser<A>& a, const parser<B>& b) { return cls<A,B>(a,b); } \
    GenBinaryOpsTokenCast(cls, op, QBtoken) \
    GenBinaryOpsTokenCast(cls, op, char) \
    GenBinaryOpsTokenCast(cls, op, int)

#include "parser_meat.hh"
#include "parser_operators.hh"
#include "parser_clike.hh"
#include "parser_nonterminal.hh"
#include "parser_lambda.hh"
