#include <SDL.h>

#include "gfx_gfx.hh"

static SDL_Surface* surface = 0;

GraphicScreen::GraphicScreen()
{
    SDL_Init(SDL_INIT_VIDEO);
    modenumber = -1;
}

#define P(rgb) { (rgb)>>16, ((rgb)>>8)&0xFF, (rgb)&0xFF, 0 }

static const SDL_Color monocolors[2] =
{ P(0x000000), P(0xFFFFFF) };

static const SDL_Color egacolors[16] =
{ P(0x000000), P(0x0000AA), P(0x00AA00), P(0x00AAAA),
  P(0xAA0000), P(0xAA00AA), P(0xAA55AA), P(0xAAAAAA),
  P(0x555555), P(0x5555FF), P(0x55FF55), P(0x55FFFF),
  P(0xFF5555), P(0xFF55FF), P(0xFF55FF), P(0xFFFFFF)
};
static const SDL_Color cgacolors[4] =
{ P(0x000000), P(0x00AAAA), P(0xAA00AA), P(0xAAAAAA) };
#undef P

static void PaletteTranslate(unsigned cpmode, int value, SDL_Color& result)
{
    switch((cpmode / 10) * 10)
    {
        case PM_Mono: B_ERROR(5); break;
        case PM_S10:
        {
            int from_level = (value % 3);
            int to_level   = (value / 3);
            int gray_level = (from_level + to_level) * 256 / 4;
            /* FIXME: should do blinking here... but. */
            result.r = gray_level;
            result.g = gray_level;
            result.b = gray_level;
            break;
        }
        case PM_EGA:
        {
            result = egacolors[value];
            break;
        }
        case PM_EGA64:
        {
            unsigned a = value & 7;
            unsigned b = value >> 3;
            result.r = std::min(255, egacolors[a].r + egacolors[b].r);
            result.g = std::min(255, egacolors[a].g + egacolors[b].g);
            result.b = std::min(255, egacolors[a].b + egacolors[b].b);
            break;
        }
        case PM_VGA:
        {
            result.r = ((value%256)   )*256/64;
            result.g = ((value>>8)%256)*256/64;
            result.b = ((value>>16)   )*256/64;
            break;
        }
    }
}


void GraphicScreen::SetGFXresolution(int xs,int ys, unsigned cpmode)
{
    surface = SDL_SetVideoMode(xs,ys, (cpmode % 10) == CM_RGB32 ? 32 : 8, 0);
    SDL_WM_SetCaption("QuickBASIC application", 0);

    /* Set maximum allowed value for COLOR command and the default palette */
    switch(cpmode % 10)
    {
        case CM_Mono: ncolors = 2;
            SDL_SetColors(surface, const_cast<SDL_Color*>(monocolors), 0, 2);
            break;
        case CM_CGA:  ncolors = 4;
            SDL_SetColors(surface, const_cast<SDL_Color*>(cgacolors), 0, 4);
            break;
        case CM_EGA:  ncolors = 16;
            SDL_SetColors(surface, const_cast<SDL_Color*>(egacolors), 0, 16);
            break;
        case CM_S10:  ncolors = 4;
            break;
        case CM_S13:  ncolors = 256;
        {
            SDL_Color mode13palette[256] = { { 0,0,0,0 } };
            /* MODE 13h (SCREEN 13) palette consists of:
             *   00..0F: EGA palette
             *   10..1F: Gradient from black to white
             *   20..37: Hue scale blue..magenta..red..yellow..green..cyan..(blue)
             *   38..4F: Hue scale (saturation=-1) = 49.2% mixed with white
             *   50..67: Hue scale (saturation=-2) = 71.4% mixed with white
             *   68..7F: Hue scale (saturation=0,  brightness=-1) = 44.4% brightness
             *   80..97: Hue scale (saturation=-1, brightness=-1)
             *   98..AF: Hue scale (saturation=-2, brightness=-1)
             *   B0..C7: Hue scale (saturation=0,  brightness=-2) = 25.4% brightness
             *   C8..DF: Hue scale (saturation=-1, brightness=-2)
             *   E0..F7: Hue scale (saturation=-2, brightness=-2)
             *   F8..FF: black
             */
            for(unsigned n=1; n<16; ++n) mode13palette[n] = egacolors[n];
            for(unsigned n=17; n<32; ++n) { int gray = (n-16)*255/15; Uint8 g=gray;
                                            SDL_Color c = { g,g,g, 0 };
                                            mode13palette[n] = c; }
            for(unsigned n=32, lightness=0; lightness<3; ++lightness)
                for(unsigned whiteness=0; whiteness<3; ++whiteness)
                    for(unsigned hue=0; hue<24; ++hue, +n)
                    {
                        static const double whitelevels[3] = { 0x00/63.0, 0x1F/63.0, 0x2D/63.0 };
                        static const double lightlevels[3] = { 0x3F/63.0, 0x1C/63.0, 0x10/63.0 };
                        double white = whitelevels[whiteness], light = lightlevels[lightness];

                        static const unsigned char shifts[24] =
                            { 0,1,2,3, 4,4,4,4, 4,4,4,4,
                              4,3,2,1, 0,0,0,0, 0,0,0,0 };

                        double rshift = (shifts[(hue+0) %24] * (1-white) + 4 * white) * light;
                        double gshift = (shifts[(hue+16)%24] * (1-white) + 4 * white) * light;
                        double bshift = (shifts[(hue+8) %24] * (1-white) + 4 * white) * light;

                        mode13palette[n].r = (int)(rshift * 255 / 4.0 + 0.5);
                        mode13palette[n].g = (int)(rshift * 255 / 4.0 + 0.5);
                        mode13palette[n].b = (int)(rshift * 255 / 4.0 + 0.5);
                    }

            SDL_SetColors(surface, mode13palette, 0, 256);
            break;
        }
        case CM_RGB32: ncolors = ~0u; break;
    }

    /* Set maximum allowed value for PALETTE command */
    switch((cpmode / 10) * 10)
    {
        case PM_Mono: npalette = 0; break;
        case PM_S10:  npalette = 9; break;
        case PM_EGA:  npalette = 16; break;
        case PM_EGA64: npalette = 64; break;
        case PM_VGA: npalette = 64*64*64; break;
    }
}

GraphicScreen::~GraphicScreen()
{
    SDL_QuitSubSystem(SDL_INIT_VIDEO);
}


void GraphicScreen::SetPalette(int index, int value)
{
    if(index >= ncolors || value >= npalette || index < 0 || value < 0) B_ERROR(5);
    SDL_Color color;
    PaletteTranslate(cpmode, value, color);
    SDL_SetColors(surface, &color, index, 1);
}

void GraphicScreen::Hline(int x1,int x2,int y,unsigned color)
{
    const std::pair<std::pair<int,int>, std::pair<int,int> >& view = GetGFXview();

    if((cpmode % 10) == CM_RGB32)
    {
        uint32_t* target = (uint32_t*)surface->pixels;
        for(int x=x1; x<=x2; ++x)
            target[y * GetGFXresolution().first + x] = color;
    }
    else
    {
        unsigned char* target = (unsigned char*)surface->pixels;
        for(int x=x1; x<=x2; ++x)
            target[y * GetGFXresolution().first + x] = color;
    }
}

void GraphicScreen::Vline(int y1,int y2,int x,unsigned color)
{
    const std::pair<std::pair<int,int>, std::pair<int,int> >& view = GetGFXview();

    if((cpmode % 10) == CM_RGB32)
    {
        uint32_t* target = (uint32_t*)surface->pixels;
        for(int y=y1; y<=y2; ++y)
            target[y * GetGFXresolution().first + x] = color;
    }
    else
    {
        unsigned char* target = (unsigned char*)surface->pixels;
        for(int y=y1; y<=y2; ++y)
            target[y * GetGFXresolution().first + x] = color;
    }
}

unsigned GraphicScreen::GetPixel(int x,int y)
{
    const std::pair<std::pair<int,int>, std::pair<int,int> >& view = GetGFXview();

    if(x < view.first.first || y < view.first.second
    || x >= view.second.first || y >= view.second.second) return 0;

    if((cpmode % 10) == CM_RGB32)
    {
        uint32_t* target = (uint32_t*)surface->pixels;
        return target[y * GetGFXresolution().first + x];
    }
    else
    {
        unsigned char* target = (unsigned char*)surface->pixels;
        return target[y * GetGFXresolution().first + x];
    }
}

void GraphicScreen::SetPixel(int x,int y,unsigned color)
{
    const std::pair<std::pair<int,int>, std::pair<int,int> >& view = GetGFXview();

    if(x < view.first.first || y < view.first.second
    || x >= view.second.first || y >= view.second.second) return;

    if((cpmode % 10) == CM_RGB32)
    {
        uint32_t* target = (uint32_t*)surface->pixels;
        target[y * GetGFXresolution().first + x] = color;
    }
    else
    {
        unsigned char* target = (unsigned char*)surface->pixels;
        target[y * GetGFXresolution().first + x] = color;
    }
}

void GraphicScreen::ScreenPOKE(int x,int y, int, int value)
{
    if(cpmode != CM_RGB32)
    {
        unsigned char* target = (unsigned char*)surface->pixels;
        target[y * GetGFXresolution().first + x] = value;
    }
}

void GraphicScreen::ClearEntireScreen()
{
    auto p = GetGFXresolution();
    unsigned num_pixels = p.first * p.second;
    unsigned color = GetBGC();
    if((cpmode % 10) == CM_RGB32)
    {
        uint32_t* target = (uint32_t*)surface->pixels;
        for(unsigned a=0; a<num_pixels; ++a)
            target[a] = color;
    }
    else
    {
        unsigned char* target = (unsigned char*)surface->pixels;
        for(unsigned a=0; a<num_pixels; ++a)
            target[a] = color;
    }
}
void GraphicScreen::ClearText(int firstline, int lastline)
{
    unsigned text_height = 8; // FIXME

    int y1 = (firstline-1) * text_height;
    int y2 = lastline * text_height - 1;
    int width = GetGFXresolution().first;
    unsigned color = GetBGC();

    for(int y = y1; y < y2; ++y)
        Hline(0, width-1, y, color);
}
void GraphicScreen::SetPages(int npages)
{
    // TODO
}
void GraphicScreen::SetAPage(int pageno)
{
    // TODO
}
void GraphicScreen::SetVPage(int pageno)
{
    // TODO
}
std::pair<int,int> GraphicScreen::GetCursorPos()
{
    // TODO
}
void GraphicScreen::SetCursorPos(const std::pair<int,int>& newpos)
{
    // TODO
}
void GraphicScreen::SetCursorSize(int start, int stop)
{
    // TODO
}
void GraphicScreen::SetCursorVisibility(bool visible)
{
    // TODO
}

std::pair<int,int> GraphicScreen::GetWinSize()
{
    // TODO
}
void GraphicScreen::SetWinSize(const std::pair<int,int>& newsize, bool approximate)
{
    // TODO
}

void GraphicScreen::SetBorderColor(int border)
{
    // TODO
}

void GraphicScreen::OutS(const char* buffer, size_t nbytes)
{
    // TODO
}

std::pair<int,int> GraphicScreen::ReadScreen(int x,int y)
{
    return {0x20, 0x20};
}


// SDL_Flip(surface)
// unsigned char* target = (unsigned char*)surface->pixels;
