#include <cstdio>
#include <cstring>
#include <memory>

#include <unistd.h> // isatty
#include <fcntl.h>    // O_RDONLY
#include <sys/stat.h>
#include <sys/file.h> // flock

#include <assert.h>

#include "stream.hh"
#include "error.hh"
#include "input.hh"
#include "gfx.hh" 

#include <map>

/* File writing */
class QbFPstream: public PrintStreamBase
{
public:
    FILE* fp;
    int xpos;
    std::vector<char> buffer;
public:
    QbFPstream(int fd, unsigned reclen=0) : fp(fd < 0 ? nullptr : fdopen(fd, "w")), xpos(0)
    {
        if(fd < 0) return;
        if(!fp) std::perror("fdopen(\"w\")");
        else if(reclen) { buffer.resize(reclen); std::setvbuf(fp, &buffer[0], _IOFBF, reclen); }
    }

    virtual unsigned GetXpos() const
    {
        return xpos;
    }
    virtual unsigned GetXsize() const
    {
        return 65535;
    }
    virtual void DoCRLF()
    {
        fputc('\n', fp);
        fflush(fp);
    }
    virtual void Write(const void* buf, size_t n)
    {
        size_t res = fwrite(buf, 1, n, fp);
        if(!res) B_ERROR_Errno();
        fflush(fp);
    }
};

/* File input */
class QbFIstream: public InputImplement
{
public:
    FILE* fp;
    std::vector<char> buffer;
public:
    QbFIstream(int fd, unsigned reclen=0) : fp(fd < 0 ? nullptr : fdopen(fd, "r"))
    {
        if(fd < 0) return;
        if(!fp) std::perror("fdopen(\"r\")");
        else if(reclen) { buffer.resize(reclen); std::setvbuf(fp, &buffer[0], _IOFBF, reclen); }
    }

    virtual void ReadLine(BAS_STR& target)
    {
        target = "";
        for(;;)
        {
            int c = InputChar();
            if(c == '\r') continue;
            if(c == '\n' || c == -1) break;
            target += (char) c;
        }
    }

    virtual void BeginInput()
    {
        InputError = false;
        GotEOL     = false;
        // File:
        //   Do nothing. InputTo reads file.
    }

    virtual bool EndInputOkp()
    {
        // File:
        //   Return false if InputTo failed, true otherwise.
        if(InputError) return false;
        return !FinishInput();
    }

    virtual BAS_STR Input_N(size_t n)
    {
        BAS_STR buf(n, '\0');
        size_t res = fread(&buf[0], 1, n, fp);
        if(res == 0)
        {
            if(ferror(fp)) { clearerr(fp); B_ERROR_Errno(); }
        }
        buf.erase(buf.begin() + res, buf.end());
        return buf;
    }

    virtual int InputChar()
    {
     retry:;
        int res = fgetc(fp);
        if(res == '\r') goto retry;
        if(res == EOF || res == '\n') { GotEOL = true; return -1; }
        return res;
    }

    virtual void GotOverflow()
    {
        B_ERROR(6); // overflow
        InputError = true;
    }
};

/* Console I/O */
class QbConsoleStream: public PrintStreamBase, public InputImplement
{
    BAS_STR InputLineBuffer;
    unsigned InputLineBufferPos;
public:
    QbConsoleStream() : PrintStreamBase(), InputImplement() { }

    virtual unsigned GetXpos() const
    {
        return B_FuncPOS();
    }
    virtual unsigned GetXsize() const
    {
        return B_FuncWIDTH();
    }
    virtual void DoCRLF()
    {
        GFX_OutS("\r", 1);
    }
    virtual void Write(const void* buf, size_t n)
    {
        GFX_OutS( (const char*) buf, n);
    }

    virtual void ReadLine(BAS_STR& target)
    {
        target = "";
        int strpos  = 0;
        int curspos = B_FuncPOS();
        bool insertmode = true;
        for(;;)
        {
            BAS_STR inkey = B_FuncINKEY_S(true);
            if(inkey == "\n" || inkey == "\r") break;
            if(inkey[0] == '\0')
                switch(inkey[1])
                {
                    case 'K': // left
                        if(strpos > 0) { --strpos; --curspos; goto relocate; }
                        break;
                    case 'M': // right
                        if(strpos < target.size())
                                       { ++strpos; ++curspos; goto relocate; }
                        break;
                    case 'O': // end
                        curspos += target.size() - strpos;
                        strpos = target.size(); goto relocate;
                    case 'Q': // home
                        curspos -= strpos; strpos = 0; goto relocate;
                    default: continue;
                }
            else if(inkey.size() == 1)
                switch(inkey[0])
                {
                    case 8: // backspace
                        if(strpos > 0)
                            { target.erase(--strpos, 1); --curspos;
                              B_StmtLOCATE(-1, curspos);
                              GFX_OutS(&target[strpos], target.size()-strpos);
                              GFX_OutS(" ", 1);
                              goto relocate;
                            }
                        continue;
                }

            target.insert(strpos, inkey);

            GFX_OutS(&target[strpos], target.size()-strpos);

            if(insertmode)
            {
                strpos += inkey.size();
                curspos += inkey.size();
            }
            else
                target.erase(strpos + inkey.size(), inkey.size());
        relocate:
            B_StmtLOCATE(-1, curspos);
        }

        if(strpos != target.size())
        {
            curspos += target.size() - strpos;
            strpos = target.size();
            B_StmtLOCATE(-1, curspos);
        }
        // note: does not go to next line.
    }

    virtual void BeginInput()
    {
        InputError = false;
        GotEOL     = false;
        // Console:
        //   Read line into buffer. InputTo queries buffer.
        ReadLine(InputLineBuffer); // note: keeps cursor on same line
        InputLineBufferPos=0;
    }

    virtual bool EndInputOkp()
    {
        // Console:
        //   Return false if InputTo failed.
        //   Return false if buffer is not empty.
        //   Return true otherwise.

        if(InputError) return false;
        return !FinishInput();
    }

    virtual BAS_STR Input_N(size_t n)
    {
        // No echoing
        // FIXME: Don't take control chars.
        BAS_STR res;
        for(; n > 0; --n) res += B_FuncINKEY_S(true);
        return res;
    }

    virtual int InputChar()
    {
        if(InputLineBufferPos < InputLineBuffer.size())
        {
            return InputLineBuffer[InputLineBufferPos++];
        }
        GotEOL = true;
        return -1;
    }

    virtual void GotOverflow()
    {
        InputError = true;
        *this << "Overflow\n";
    }
};

struct FileDescriptor
{
    int fd; // system filedescriptor
    QbFPstream fpout;
    QbFIstream fpin;

    std::size_t record_width;
    std::size_t record_offset;

    struct Field
    {
        std::size_t offset;
        std::size_t width;
        std::string* target;
    };

    std::vector<Field> fields;

    bool readable;
    bool writable;

public:
    FileDescriptor(int f, unsigned reclen, bool r,bool w)
        : fd(f),
          fpout(w ? f : -1, reclen),
          fpin(r ? f : -1, reclen),
          record_width(reclen), record_offset(0),
          fields(), readable(r), writable(w)
    {
    }
};
static std::map<int/*qbasic filenum*/, std::shared_ptr<FileDescriptor>> fdmap;

PrintStreamBase& GET_STREAM_OUTPUT(int filenumber)
{
    auto i = fdmap.find(filenumber);
    if(i == fdmap.end()) B_ERROR(52); // bad file number
    if(!i->second->writable) B_ERROR(54); // bad file mode

    return i->second->fpout;
}

InputStreamBase& GET_STREAM_INPUT(int filenumber)
{
    auto i = fdmap.find(filenumber);
    if(i == fdmap.end()) B_ERROR(52); // bad file number
    if(!i->second->readable) B_ERROR(54); // bad file mode
    return i->second->fpin;
}

static QbConsoleStream cons;

PrintStreamBase& GET_STDOUT()
{
    static bool tty_out = isatty(1);
    if(!tty_out) { static QbFPstream out(1); return out; }
    return cons;
}
InputStreamBase& GET_STDIN()
{
    static bool tty_in = isatty(0);
    if(!tty_in) { static QbFIstream in(0); return in; }
    return cons;
}

BAS_INT B_FuncFREEFILE()
{
    if(fdmap.empty()) return 1;
    return fdmap.rbegin()->first + 1;
}

void B_StmtFIELD(BAS_INT filenumber, unsigned width, std::string& stringvar)
{
    auto i = fdmap.find(filenumber);
    if(i == fdmap.end()) B_ERROR(52); // bad file number
    //if(!i->second.record_width) B_ERROR(54); // bad file mode
    auto& r = *i->second;

    if(r.record_offset + width > r.record_width) B_ERROR(50); // field overflow

    stringvar.assign(width, ' ');
    r.fields.emplace_back( FileDescriptor::Field{r.record_offset, width, &stringvar} );
    r.record_offset += width;
}

void B_StmtFIELD(BAS_INT filenumber)
{
    auto i = fdmap.find(filenumber);
    if(i == fdmap.end()) B_ERROR(52); // bad file number
    auto& r = *i->second;

    r.record_offset = 0;
}

void B_StmtCLOSE()
{
    for(const auto& f: fdmap) close(f.second->fd);
    fdmap.clear();
}
void B_StmtCLOSE(BAS_INT filenumber)
{
    auto i = fdmap.find(filenumber);
    if(i == fdmap.end()) B_ERROR(52); // bad file number
    auto& r = *i->second;

    close(r.fd);
    fdmap.erase(i);
}

void B_OPEN(const BAS_STR& filename,BAS_INT filenumber,  unsigned mode, unsigned bitmask, BAS_LNG reclen)
{
    B_OPEN(filename.c_str(), filenumber, mode, bitmask, reclen);
}

void B_OPEN(const char* filename, BAS_INT filenumber,  unsigned mode, unsigned bitmask, BAS_LNG reclen)
{
    if(fdmap.find(filenumber) != fdmap.end()) B_ERROR(55); // file already open
    if(reclen < 0) B_ERROR(63); // bad record length
    if(mode < 1 || mode > 4 || (bitmask&3) == 0) B_ERROR(54); // bad file mode

    bool access_read  = bitmask&1;
    bool access_write = bitmask&2;
    bool lock_read    = bitmask&4;
    bool lock_write   = bitmask&8;

    // modes:
    //   1=RANDOM
    //   2=INPUT and OUTPUT
    //   3=BINARY
    //   4=APPEND

    bool create   = access_write;
    bool truncate = access_write && !access_read && mode!=4;
    bool append   = access_write && mode==4;

    unsigned open_flags = access_read ? access_write ? O_RDWR : O_RDONLY : O_WRONLY;
    if(create)    open_flags |= O_CREAT;
    if(truncate)  open_flags |= O_TRUNC;
    if(append)    open_flags |= O_APPEND;

    if(reclen == 0) reclen = mode==2 ? 512 : 128;
    if(mode == 3) reclen = 1; // For BINARY files, ignore the record length

    int fd = open(filename, open_flags, 0644);
    if(fd < 0) B_ERROR_Errno();

    int lock_status = flock(fd, (lock_write || lock_read) ? LOCK_EX : LOCK_SH);
    if((lock_read || lock_write) && lock_status < 0)
    {
        close(fd);
        B_ERROR_Errno();
    }
    if(fd >= 0)
    {
        fdmap.emplace(filenumber, std::make_shared<FileDescriptor>(fd,reclen, access_read,access_write));
    }
}

void B_OPEN(const BAS_STR& filename,BAS_INT filenumber, const BAS_STR& mode, unsigned bitmask, BAS_LNG reclen)
{
    if(mode.size() == 1)
        switch(mode[0])
        {
            case 'O': B_OPEN(filename,filenumber, 2, bitmask|2, reclen); return;
            case 'I': B_OPEN(filename,filenumber, 2, bitmask|1, reclen); return;
            case 'R': B_OPEN(filename,filenumber, 1, bitmask|3, reclen); return;
            case 'B': B_OPEN(filename,filenumber, 3, bitmask|3, reclen); return;
            case 'A': B_OPEN(filename,filenumber, 4, bitmask|2, reclen); return;
        }
    B_ERROR(54); // bad file mode
}

void B_StmtGET(BAS_INT filenumber, BAS_LNG recnumber)
{
    auto i = fdmap.find(filenumber);
    if(i == fdmap.end()) B_ERROR(52); // bad file number
    auto& r = *i->second;
    if(!r.readable) B_ERROR(54);      // bad file mode
    if(r.fields.empty()) B_ERROR(40); // variable required

    // TODO: If file is open in INPUT mode, "bad file mode"
    if(recnumber < -1 || recnumber == 0) B_ERROR(63); // bad record number

    if(recnumber > 0) std::fseek(r.fpin.fp, (recnumber-1) * r.record_width, SEEK_SET);

    std::vector<char> buffer(r.record_width);

    if(std::fread(&buffer[0], 1, buffer.size(), r.fpin.fp) == 0)
    {
        if(std::ferror(r.fpin.fp))
            B_ERROR_Errno();
        else
            B_ERROR(62); // input past end of file
    }

    for(const auto& field: r.fields)
    {
        assert(field.offset + field.width <= buffer.size());
        assert(field.width <= field.target->size());
        field.target->assign( &buffer[field.offset], field.width );
    }
}

void B_StmtGET(BAS_INT filenumber, BAS_LNG recnumber, void* target, std::size_t bytes)
{
    auto i = fdmap.find(filenumber);
    if(i == fdmap.end()) B_ERROR(52); // bad file number
    auto& r = *i->second;
    if(!r.readable) B_ERROR(54);      // bad file mode

    if(!r.fields.empty()) B_ERROR(56); // field statement active

    // TODO: If file is open in INPUT or RANDOM mode, "bad file mode"

    if(recnumber < -1 || recnumber == 0) B_ERROR(63); // bad record number
    if(recnumber > 0) std::fseek(r.fpin.fp, (recnumber-1) * r.record_width, SEEK_SET);

    // TODO: If file is open in RANDOM mode and bytes doesn't match record_width, error

    if(std::fread(target, 1, bytes, r.fpin.fp) == 0)
    {
        if(std::ferror(r.fpin.fp))
            B_ERROR_Errno();
        else
            B_ERROR(62); // input past end of file
    }
}

void B_StmtPUT(BAS_INT filenumber, BAS_LNG recnumber)
{
    auto i = fdmap.find(filenumber);
    if(i == fdmap.end()) B_ERROR(52); // bad file number
    auto& r = *i->second;
    if(!r.writable) B_ERROR(54);      // bad file mode
    if(r.fields.empty()) B_ERROR(40); // variable required

    // TODO: If file is open in OUTPUT mode, "bad file mode"
    if(recnumber < -1 || recnumber == 0) B_ERROR(63); // bad record number
    if(recnumber > 0) std::fseek(r.fpout.fp, (recnumber-1) * r.record_width, SEEK_SET);

    std::vector<char> buffer(r.record_width);

    for(const auto& field: r.fields)
    {
        assert(field.offset + field.width <= buffer.size());
        assert(field.width <= field.target->size());
        std::copy( field.target->begin(),
                   field.target->begin() + field.width,
                   buffer.begin() + field.offset );
    }
    if(std::fwrite(&buffer[0], 1, buffer.size(), r.fpout.fp) == 0) B_ERROR_Errno();
}

void B_StmtPUT(BAS_INT filenumber, BAS_LNG recnumber, const void* source, std::size_t bytes)
{
    auto i = fdmap.find(filenumber);
    if(i == fdmap.end()) B_ERROR(52); // bad file number
    auto& r = *i->second;
    if(!r.writable) B_ERROR(54);      // bad file mode

    if(!r.fields.empty()) B_ERROR(56); // field statement active

    // TODO: If file is open in INPUT or RANDOM mode, "bad file mode"

    if(recnumber < -1 || recnumber == 0) B_ERROR(63); // bad record number
    if(recnumber > 0) std::fseek(r.fpout.fp, (recnumber-1) * r.record_width, SEEK_SET);

    // TODO: If file is open in RANDOM mode and bytes doesn't match record_width, error

    if(std::fwrite(source, 1, bytes, r.fpout.fp) == 0) B_ERROR_Errno();
}
