#ifndef bqtQBlibStreamHH
#define bqtQBlibStreamHH

#include <qblib/types.hh>

#include <sstream>
#include <string>
#include <cstddef> // for size_t
#include <vector>

#include <stdint.h>

struct PrintCRLFtag { };
struct PrintSPCtag  { int count; };
struct PrintTABtag  { int count; };
struct PrintNextTABtag { };

class PrintStreamBase
{
public:
    PrintStreamBase() {}
    virtual ~PrintStreamBase() {}

    PrintStreamBase& operator<< (const BAS_INT v);
    PrintStreamBase& operator<< (const BAS_LNG v);
    PrintStreamBase& operator<< (const BAS_SGN v);
    PrintStreamBase& operator<< (const BAS_DBL v);
    PrintStreamBase& operator<< (const BAS_STR& v);
    PrintStreamBase& operator<< (const long long int v);
    PrintStreamBase& operator<< (const char* v);
    PrintStreamBase& operator<< (const char v);

    PrintStreamBase& operator<< (PrintCRLFtag)    { DoCRLF(); return *this; }
    PrintStreamBase& operator<< (PrintNextTABtag) { DoNextTAB(); return *this; }
    PrintStreamBase& operator<< (PrintSPCtag t)   { DoSPC(t.count); return *this; }
    PrintStreamBase& operator<< (PrintTABtag t)   { DoTAB(t.count); return *this; }

    virtual void Write(const void* buf, size_t n) = 0;

    // DoCRLF is not the same as either '\r' nor '\n'.
    // Only on terminal, '\r' and '\n' both produce the same as DoCRLF().
    // In files, DoCRLF() produces '\r\n' whereas CHR$ produces individual bytes.
    virtual void DoCRLF() = 0;

    virtual unsigned GetXpos() const = 0;  // 0..n-1
    virtual unsigned GetXsize() const = 0; // n. 0 = infinite (files)

    // SPC() actually produces space characters. It does not only move the cursor.
    void DoSPC(int count);
    void DoTAB(int count);
    void DoNextTAB();
private:
    bool CheckFit(int n_columns); // no const, it may print CRLF
    void WriteWithFitCheck(const void* buf, size_t n);
    void WriteWithFitCheck(const std::string& s)
        { WriteWithFitCheck(s.data(), s.size()); }
    void Write(const std::string& s)
        { Write(s.data(), s.size()); }

private:
    PrintStreamBase(const PrintStreamBase&);
    void operator=(const PrintStreamBase&);
};

class InputStreamBase
{
public:
    InputStreamBase() {}
    virtual ~InputStreamBase() {}

    // LINE INPUT (console and file)
    virtual void ReadLine(BAS_STR& target) = 0;

    // INPUT$  (console and file)
    virtual BAS_STR Input_N(size_t n) = 0;

    virtual void BeginInput() = 0;
    virtual void InputTo(BAS_INT& target) = 0;
    virtual void InputTo(BAS_LNG& target) = 0;
    virtual void InputTo(BAS_SGN& target) = 0;
    virtual void InputTo(BAS_DBL& target) = 0;
    virtual void InputTo(BAS_STR& target) = 0;
    virtual bool EndInputOkp() = 0; // end input, were all fields input?
    /* In console input:
         At call of EndInputOkp, false is returned
         if any of the InputTos caused error or if
         line was not completely emptied by InputTos
       In file input:
         InputTo may span multiple lines, half-read
         line does not matter. EndInputOkp returns false
         only if an InputTo was unsatisfied.
     */

    static inline const char* GetRedoMsg() { return "\n?Redo from start\n"; }

private:
    InputStreamBase(const InputStreamBase&);
    void operator=(const InputStreamBase&);
};

extern PrintStreamBase& GET_STREAM_OUTPUT(int filenumber);
extern PrintStreamBase& GET_PRINTER();
extern PrintStreamBase& GET_STDOUT();

extern InputStreamBase& GET_STDIN();
extern InputStreamBase& GET_STREAM_INPUT(int filenumber);

class UsingFormatter
{
private:
    struct FormatItem
    {
        enum SignType
        {
            NormalSignLeft,    // number slot or -
            AlwaysSignLeft,    // + or -
            NegativeSignRight, // - or space (fill_char does not affect)
            AlwaysSignRight    // + or -
        };

        enum { Literal, String, Numeric } type;
        // For literals
        std::string literal;
        // For strings
        int         length; // -1=entire string
        // For numeric
        union
        {
            struct
            {
                /* Note: These fields are so constructed that 0 is a sensible
                 *       default for each field. */
                unsigned    num_leading_chars       : 8;
                unsigned    num_trailing_chars      : 8;
                unsigned    scientific              : 2; // 0, 2 or 3
                bool        fill_stars              : 1; // 0=spaces, 1=stars
                bool        decimal_point           : 1;
                bool        comma_every_third_digit : 1;
                bool        leading_dollar          : 1; // note: prints $ instead of last space
                SignType    sign_type               : 2;
                                            //       note: goes _after_ SignLeft
            };
            uint_least32_t numdata;
        };
    };
    std::vector<FormatItem> format;
    size_t pos;
public:
    explicit UsingFormatter(const BAS_STR& f);
    ~UsingFormatter();

    /* Note: DoParam() can execute B_ERROR(13) (Type mismatch) */
    PrintStreamBase& DoParam(PrintStreamBase& out, const BAS_LNG v);
    PrintStreamBase& DoParam(PrintStreamBase& out, const BAS_DBL v);
    PrintStreamBase& DoParam(PrintStreamBase& out, const BAS_STR& v);
    PrintStreamBase& DoParam(PrintStreamBase& out, const BAS_INT v) { return DoParam(out, BAS_LNG(v)); }
    PrintStreamBase& DoParam(PrintStreamBase& out, const BAS_BYTE v) { return DoParam(out, BAS_LNG(v)); }
    PrintStreamBase& DoParam(PrintStreamBase& out, const BAS_UBYTE v) { return DoParam(out, BAS_LNG(v)); }
    PrintStreamBase& DoParam(PrintStreamBase& out, const BAS_UINT v) { return DoParam(out, BAS_LNG(v)); }
    PrintStreamBase& DoParam(PrintStreamBase& out, const BAS_ULONG v);

    PrintStreamBase& DoParam(PrintStreamBase& out, const BAS_SGN v);
    PrintStreamBase& DoParam(PrintStreamBase& out, const char* v) { return DoParam(out, BAS_STR(v)); }
    PrintStreamBase& DoParam(PrintStreamBase& out, const char v) { return DoParam(out, BAS_STR(1,v)); }

    PrintStreamBase& DoParam(PrintStreamBase& out, const BAS_DBL v, char ExpChar);

    template<typename T, typename... Rest>
    inline PrintStreamBase& DoParam(PrintStreamBase& out, const T& value, Rest... rest)
    {
        return DoParam(DoParam(out, value), rest...);
    }

private:
    const FormatItem& GetFormat();
    void FlushLiterals(PrintStreamBase& out, bool allow_wrap);

    void CreateFormat(const BAS_STR& f);
    static void OutputNonscientific(
        PrintStreamBase& out,
        const BAS_DBL v,
        const UsingFormatter::FormatItem& format);

private:
    UsingFormatter(const UsingFormatter&);
    void operator=(const UsingFormatter&);
};


static inline const BAS_STR B_FuncINPUT_S(int n)                 { return GET_STDIN().Input_N(n); }
static inline const BAS_STR B_FuncINPUT_S(int n, int filenumber) { return GET_STREAM_INPUT(filenumber).Input_N(n); }

extern BAS_INT B_FuncFREEFILE();
extern void B_StmtFIELD(BAS_INT filenumber);
extern void B_StmtFIELD(BAS_INT filenumber, unsigned width, std::string& stringvar);
extern void B_StmtCLOSE();
extern void B_StmtCLOSE(BAS_INT filenumber);

// See stmt_fileopen::Compile()
extern void B_OPEN(const BAS_STR& filename,BAS_INT filenumber, unsigned mode, unsigned bitmask, BAS_LNG reclen=0);
extern void B_OPEN(const BAS_STR& filename,BAS_INT filenumber, const BAS_STR& mode, unsigned bitmask, BAS_LNG reclen=0);
extern void B_OPEN(const char* filename,BAS_INT filenumber, unsigned mode, unsigned bitmask, BAS_LNG reclen=0);

extern void B_StmtGET(BAS_INT filenumber, BAS_LNG recnumber=-1);
extern void B_StmtGET(BAS_INT filenumber, BAS_LNG recnumber, void* target, std::size_t bytes);

static inline void B_StmtGET(BAS_INT filenumber, BAS_LNG recnumber, BAS_STR& target)
    { B_StmtGET(filenumber,recnumber, target.empty() ? nullptr : (void*)&target[0], target.size()); }

template<typename T>
static inline void B_StmtGET(BAS_INT filenumber, BAS_LNG recnumber, T& target)
    { B_StmtGET(filenumber,recnumber,(void*)&target, sizeof(target)); }

extern void B_StmtPUT(BAS_INT filenumber, BAS_LNG recnumber=-1);
extern void B_StmtPUT(BAS_INT filenumber, BAS_LNG recnumber, const void* source, std::size_t bytes);

static inline void B_StmtPUT(BAS_INT filenumber, BAS_LNG recnumber, const BAS_STR& input)
    { B_StmtPUT(filenumber,recnumber, input.empty() ? nullptr : (const void*)&input[0], input.size()); }

template<typename T>
static inline void B_StmtPUT(BAS_INT filenumber, BAS_LNG recnumber, const T& source)
    { B_StmtPUT(filenumber,recnumber,(const void*)&source, sizeof(source)); }

#endif
